"""Garjus Issues Management."""
import os

import logging
import importlib


logger = logging.getLogger('garjus.issues')


def update(garjus, projects=None):
    """Update issues."""

    if not garjus.xnat_enabled():
        logger.debug('no xnat, cannot update issues')
        return

    # First find unmatched sessions across projects.
    # these are sessions that are in source project
    # but not found in destination projects
    unmatched = _unmatched(garjus)

    # TODO: only need to check projects with same source project

    # Update each project
    for p in (projects or garjus.projects()):
        if p in projects:
            logger.debug(f'updating issues:{p}')
            update_project(garjus, p, unmatched[p])


def _unmatched(garjus):
    src2dst = {}
    src2ignore = {}
    unmatched = {}

    # Build the list of dst projects (and ignore list) for each source project
    for dst_project in garjus.projects():
        unmatched[dst_project] = []

        for p in garjus.scanning_protocols(dst_project):
            src_project = p['scanning_srcproject']

            if not src_project:
                logger.debug(f'no scanning_srcproject')
                continue

            # Get sessions to ignore
            ignore_sessions = p['scanning_ignore'].split(',')
            ignore_sessions = [x.strip() for x in ignore_sessions]

            # Append dst project to list
            if src_project not in src2dst:
                # Make a new list
                src2dst[src_project] = [dst_project]
                src2ignore[src_project] = ignore_sessions

            elif dst_project not in src2dst[src_project]:
                # Add to existing lists
                src2dst[src_project].append(dst_project)
                src2ignore[src_project].extend(ignore_sessions)

    # Find unmatched in each source project
    for src_project, dst_projects in src2dst.items():
        logger.debug(f'finding unmatched sessions:{src_project}')
        src_labels = garjus.session_labels(src_project)
        src_ignore = src2ignore[src_project]

        # Build the list of src IDs for sessions in the destination projects
        srcid_list = []
        for dst_project in dst_projects:
            # These are the original session labels before being renamed
            srcid_list += garjus.session_source_labels(dst_project)

        # Apply ignore list
        src_labels = [x for x in src_labels if x not in src_ignore]

        # Get unmatched, not in list of sources in destination
        src_unmatched = [x for x in src_labels if x not in srcid_list]

        # Create an issue for each dst project, we don't know which is the dst
        for sess in src_unmatched:
            logger.debug(f'unmatched session:{sess}')
            for dst_project in dst_projects:
                unmatched[dst_project].append(sess)

    return unmatched


def update_project(garjus, project, unmatched):
    """Update project issues."""
    results = []

    # Load the project primary redcap
    project_redcap = garjus.primary(project)
    if not project_redcap:
        logger.debug('primary redcap not found, cannot update issues')
        return

    # Confirm project exists on XNAT
    if not garjus.project_exists(project):
        msg = f'destination project does not exist:{project}'
        logger.debug(msg)
        return

    # Audit edats
    for p in garjus.edat_protocols(project):
        # Load event list from comma-delimited value
        events = p['edat_events']
        events = [x.strip() for x in events.split(',')]

        # Run the audit
        results += _audit_edat(
            project_redcap,
            events,
            p['edat_rawfield'],
            p['edat_convfield'],
            p['edat_readyfield'])

    # Scanning protocols
    for p in garjus.scanning_protocols(project):
        events = p['scanning_events']
        date_field = p['scanning_datefield']
        sess_field = p['scanning_srcsessfield']
        sess_suffix = p['scanning_xnatsuffix']
        src_project = p['scanning_srcproject']

        if not date_field:
            logger.debug(f'missing date field')
            continue

        if not sess_field:
            logger.debug(f'missing sess field')
            continue

        src_labels = garjus.session_labels(src_project)
        dst_labels = garjus.session_labels(project)

        # Get events list
        events = None
        if p.get('scanning_events', False):
            events = [x.strip() for x in p['scanning_events'].split(',')]

        # check that projects exist on XNAT
        if not garjus.source_project_exists(src_project):
            msg = f'source project does not exist in XNAT:{src_project}'
            logger.error(msg)
            garjus.add_issue(msg, project, category='ERROR')
            continue

        # Check for alternate redcap
        if p['scanning_altprimary'] != '':
            _rc = garjus.alternate(p['scanning_altprimary'])
        else:
            _rc = project_redcap

        # Make the scan table linking scan ID to subject/session
        scan_table = _make_scan_table(
            _rc,
            events,
            date_field,
            sess_field,
            sess_suffix)

        #if project == 'DSCHOL':
        #    scan_table = _dschol_scan_table(scan_table)

        results += _audit_scanning(scan_table, src_labels, dst_labels)

    # Issues for unmatched
    for sess in unmatched:
        results.append({
            'project': project,
            'category': 'UNMATCHED_SESSION',
            'session': sess,
            'description': sess})

    # Check for externally uploaded images
    results += _audit_inbox(garjus, project)

    _add_issues(garjus, results, project)


def _matching_issues(issue1, issue2):
    # Matching means both issues are of the same category
    # and on the same Project/Subject
    # and as applicable, the same XNAT Session/Scan
    # and as applicable the same REDCap Event/Field
    keys = [
        'PROJECT', 'CATEGORY', 'SUBJECT', 'SESSION', 'SCAN', 'EVENT', 'FIELD']

    for k in keys:
        if (k.lower() in issue1) and (issue1[k.lower()] != issue2[k]):
            return False

    # Everything matched
    return True


def _audit_edat(project, events, rawfield, convfield, readyfield):
    try:
        audit_edat = importlib.import_module(f'garjus.issues.audit_edat')
    except ModuleNotFoundError as err:
        logger.error(f'error loading:{err}')
        return

    return audit_edat.audit(project, events, rawfield, convfield, readyfield)


def _audit_scanning(scan_table, src_project, project):
    try:
        audit_imaging = importlib.import_module(f'garjus.issues.audit_imaging')
    except ModuleNotFoundError as err:
        logger.error(f'error loading:{err}')
        return

    return audit_imaging.audit(scan_table, src_project, project)


def _add_issues(garjus, records, project):
    new_issues = []
    has_errors = False

    # First check existing issues,
    # import new issues and update existing,
    # complete any no longer found

    # Check for errors
    for r in records:
        if r['category'] == 'ERROR':
            has_errors = True
            break

    # Set project
    for r in records:
        r['project'] = project

    # Compare to current issues
    cur_issues = garjus.issues(project)
    new_issues = _find_new(cur_issues, records)

    # Upload new records
    if new_issues:
        logger.debug(f'uploading {len(new_issues)} new issues')
        garjus.add_issues(new_issues)
    else:
        logger.debug('no new issues to upload')

    if has_errors:
        logger.debug(f'errors during audit, not closing old issues')
        return

    # Close fixed issues
    fixed_issues = _find_fixed(cur_issues, records)
    if fixed_issues:
        logger.debug(f'setting {len(fixed_issues)} issues to complete')
        garjus.close_issues(fixed_issues)
    else:
        logger.debug('no resolved issues to complete')


def _find_new(issues, records):
    results = []

    for rec in records:
        isnew = True

        # Try to find a matching record
        for cur in issues.to_dict('records'):
            cur_id = cur['ID']
            cur_proj = cur['PROJECT']
            if _matching_issues(rec, cur):
                isnew = False
                logger.debug(f'matches existing issue:{cur_proj}:{cur_id}')
                break

        if isnew:
            results.append(rec)

    return results


def _find_fixed(issues, records):
    """Return issues that are not in current search, i.e. resolved"""
    results = []
    # Find old issues
    logger.debug('checking for resolved issues')
    for cur in issues.to_dict('records'):
        isold = True
        cur_id = cur['ID']
        cur_proj = cur['PROJECT']

        # Try to find a matching record
        for rec in records:
            if _matching_issues(rec, cur):
                isold = False
                logger.debug(f'matches existing issue:{cur_proj}:{cur_id}')
                break

        if isold:
            # Append to list as closed with current time
            logger.debug(f'found resolved issue:{cur_proj}:{cur_id}')
            results.append({'project': cur_proj, 'id': cur_id})

    return results


def _make_scan_table(
    project,
    events,
    date_field,
    sess_field,
    scan_suffix,
):
    """Make the scan table, linking source to destination subject/session."""
    data = []
    id2subj = {}

    # Shortcut
    def_field = project.def_field

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}

    # Get mri records from redcap
    fields = [date_field, sess_field, def_field]
    try:
        rec = project.export_records(fields=fields, events=events)
    except Exception as err:
        logger.error(f'cannot make scan table:{err}:{fields}')
        return []

    # Only if date is entered
    rec = [x for x in rec if x[date_field]]

    # Only if session id entered
    rec = [x for x in rec if x[sess_field]]

    # Set the subject and session
    for r in rec:
        d = {}
        d['scandate'] = r[date_field]
        d['src_session'] = r[sess_field]
        d['src_subject'] = d['src_session']
        d['dst_subject'] = id2subj.get(r[def_field], r[def_field])
        d['dst_session'] = d['dst_subject'] + scan_suffix
        if 'redcap_event_name' in r:
            d['event'] = r['redcap_event_name']
        else:
            d['event'] = ''

        data.append(d)

    return data


def _dschol_scan_table(data):

    for d in data:
        if d['dst_subject'].startswith('1'):
            d['dst_subject'] = 'DST30500' + d['dst_subject'][1:]
            d['dst_session'] = 'DST30500' + d['dst_session'][1:]
        elif d['dst_subject'].startswith('2'):
            d['dst_subject'] = 'DSCHOL' + d['dst_subject']
            d['dst_session'] = 'DSCHOL' + d['dst_session']

    return data


def _audit_inbox(garjus, project):
    results = []

    # Create a single issue if there are any files other than ARCHIVE
    project_inbox = garjus.project_setting(project, 'inbox')

    if not project_inbox:
        logger.debug(f'no project_inbox:{project}')
    elif not os.path.isdir(project_inbox):
        logger.error(f'cannot check, file not found:{project}:{project_inbox}')
        results.append({'category': 'ERROR','description': 'inbox not found'})
    else:
        logger.debug(f'auditing inbox:{project}:{project_inbox}')
        inbox_files = os.listdir(project_inbox)
        inbox_files = [x for x in inbox_files if x not in ['ARCHIVED', '.DS_Store']]
        inbox_files = [x for x in inbox_files if not x.endswith('.docx')]
        if len(inbox_files) > 0:
            results.append({
                'project': project,
                'category': 'UNMATCHED_SESSION',
                'description': 'Inbox contains unmatched uploads',
                'field': 'project_inbox'})

    return results



import logging
import tempfile

import pandas as pd

from ...utils_redcap import get_redcap, download_file, field2events


logger = logging.getLogger('garjus.automations.etl_nihtoolbox_drnair')


reg_field = 'toolbox_regdata'
score_field = 'toolbox_cogscores'
done_field = 'toolbox_pin'


def process(project):
    results = []
    events = field2events(project, reg_field)

    records = project.export_records(
        fields=[project.def_field, done_field, reg_field, score_field],
        events=events)

    for r in records:
        record_id = r[project.def_field]
        event_id = r['redcap_event_name']

        if r[done_field]:
            logger.debug(f'already ETL:{record_id}:{event_id}')
            continue

        if not r[reg_field]:
            logger.debug(f'no reg file:{record_id}:{event_id}')
            continue

        if not r[score_field]:
            logger.debug(f'no data file:{record_id}:{event_id}')
            continue

        logger.debug(f'running ETL:{record_id}:{event_id}')
        results.append({'subject': record_id, 'event': event_id})
        _run(project, record_id, event_id)

    return results


def _run(project, record_id, event_id):
    data = None

    with tempfile.TemporaryDirectory() as tmpdir:
        reg_file = f'{tmpdir}/regfile.csv'
        score_file = f'{tmpdir}/scorefile.csv'

        # Download files from redcap
        logger.debug(f'downloading file:{record_id}:{event_id}:{reg_field}:{reg_file}')
        download_file(project, record_id, reg_field, reg_file, event_id=event_id)
        logger.debug(f'downloading file:{record_id}:{event_id}:{score_field}:{score_file}')
        download_file(project, record_id, score_field, score_file, event_id=event_id)

        # Extract data from downloaded files
        reg_data = _extract_regdata(reg_file)

        score_data = _extract_cogscores(score_file)

        # Transform data to match redcap field names
        data = _transform(reg_data, score_data)

    # Load data back to redcap
    print(data)
    _load(project, record_id, event_id, data)


def _transform(regdata, scoredata):
    # Initialize test data
    picseqtest = {}
    listsorttest = None
    patterntest = None
    picvocabtest = None
    oralrecogtest = None
    cogcrystalcomp = None
    data = {}

    # Start with the registration data
    data.update({
        'toolbox_pin': regdata['PIN'],
        'toolbox_deviceid': regdata['DeviceID'],
        'toolbox_age': regdata['Age'],
        'toolbox_education': regdata['Education'],
        'toolbox_gender': regdata['Gender'],
        'toolbox_handedness': regdata['Handedness'],
        'toolbox_race': regdata['Race'],
        'toolbox_ethnicity': regdata['Ethnicity'],
        'toolbox_assessment': regdata['Assessment Name'],
    })

    # Find the Pic Seq data that has mutliple versions
    for i in list(scoredata.keys()):
        if i.startswith('NIH Toolbox Picture Sequence Memory Test'):
            picseqtest = scoredata[i]

    # Load the other tests
    listsorttest = scoredata.get('NIH Toolbox List Sorting Working Memory Test Age 7+ v2.1', {})
    patterntest = scoredata.get('NIH Toolbox Pattern Comparison Processing Speed Test Age 7+ v2.1', {})
    picvocabtest = scoredata.get('NIH Toolbox Picture Vocabulary Test Age 3+ v2.1', {})
    oralrecogtest = scoredata.get('NIH Toolbox Oral Reading Recognition Test Age 3+ v2.1', {})
    flankertest = scoredata.get('NIH Toolbox Flanker Inhibitory Control and Attention Test Age 12+ v2.1', {})
    cardsorttest = scoredata.get('NIH Toolbox Dimensional Change Card Sort Test Age 12+ v2.1', {})

    # Get the individual scores
    data.update({
        'toolbox_listsorttest_raw': listsorttest.get('RawScore', ''),
        'toolbox_patterntest_raw': patterntest.get('RawScore', ''),
        'toolbox_picseqtest_raw': picseqtest.get('RawScore', ''),
        'toolbox_oralrecogtest_theta': oralrecogtest.get('Theta', ''),
        'toolbox_picseqtest_theta': picseqtest.get('Theta', ''),
        'toolbox_picvocabtest_theta': picvocabtest.get('Theta', ''),
        'toolbox_listsorttest_uncstd': listsorttest.get('Uncorrected Standard Score', ''),
        'toolbox_oralrecogtest_uncstd': oralrecogtest.get('Uncorrected Standard Score', ''),
        'toolbox_patterntest_uncstd': patterntest.get('Uncorrected Standard Score', ''),
        'toolbox_picseqtest_uncstd': picseqtest.get('Uncorrected Standard Score', ''),
        'toolbox_picvocabtest_uncstd': picvocabtest.get('Uncorrected Standard Score', ''),
        'toolbox_listsorttest_agestd': listsorttest.get('Age-Corrected Standard Score', ''),
        'toolbox_oralrecogtest_agestd': oralrecogtest.get('Age-Corrected Standard Score', ''),
        'toolbox_patterntest_agestd': patterntest.get('Age-Corrected Standard Score', ''),
        'toolbox_picseqtest_agestd': picseqtest.get('Age-Corrected Standard Score', ''),
        'toolbox_picvocabtest_agestd': picvocabtest.get('Age-Corrected Standard Score', ''),
        'toolbox_listsorttest_tscore': listsorttest.get('Fully-Corrected T-score', ''),
        'toolbox_oralrecogtest_tscore': oralrecogtest.get('Fully-Corrected T-score', ''),
        'toolbox_patterntest_tscore': patterntest.get('Fully-Corrected T-score', ''),
        'toolbox_picseqtest_tscore': picseqtest.get('Fully-Corrected T-score', ''),
        'toolbox_picvocabtest_tscore': picvocabtest.get('Fully-Corrected T-score', ''),
        'toolbox_flankertest_raw': flankertest.get('RawScore', ''),
        'toolbox_flankertest_uncstd': flankertest.get('Uncorrected Standard Score', ''),
        'toolbox_flankertest_agestd': flankertest.get('Age-Corrected Standard Score', ''),
        'toolbox_flankertest_tscore': flankertest.get('Fully-Corrected T-score', ''),
        'toolbox_cardsorttest_raw': cardsorttest.get('RawScore', ''),
        'toolbox_cardsorttest_uncstd': cardsorttest.get('Uncorrected Standard Score', ''),
        'toolbox_cardsorttest_agestd': cardsorttest.get('Age-Corrected Standard Score', ''),
        'toolbox_cardsorttest_tscore': cardsorttest.get('Fully-Corrected T-score', ''),
    })

    cogcrystalcomp = scoredata.get('Cognition Crystallized Composite v1.1', None)
    cogfluidcomp = scoredata.get('Cognition Fluid Composite v1.1', None)
    cogearlycomp = scoredata.get('Cognition Early Childhood Composite v1.1', None)
    cogtotalcomp = scoredata.get('Cognition Total Composite Score v1.1', None)
    audlearntest = scoredata.get('NIH Toolbox Auditory Verbal Learning Test (Rey) Age 8+ v2.0', None)

    if audlearntest:
        data.update({
            'toolbox_audlearntest_raw': audlearntest['RawScore'],
        })

    if cogcrystalcomp:
        data.update({
            'toolbox_cogcrystalcomp_uncstd': cogcrystalcomp['Uncorrected Standard Score'],
            'toolbox_cogcrystalcomp_agestd': cogcrystalcomp['Age-Corrected Standard Score'],
            'toolbox_cogcrystalcomp_tscore': cogcrystalcomp['Fully-Corrected T-score'],
        })

    if cogfluidcomp:
        data.update({
            'toolbox_cogfluidcomp_uncstd': cogfluidcomp['Uncorrected Standard Score'],
            'toolbox_cogfluidcomp_agestd': cogfluidcomp['Age-Corrected Standard Score'],
            'toolbox_cogfluidcomp_tscore': cogfluidcomp['Fully-Corrected T-score'],
        })

    if cogearlycomp:
        data.update({
            'toolbox_cogearlycomp_uncstd': cogearlycomp['Uncorrected Standard Score'],
            'toolbox_cogearlycomp_agestd': cogearlycomp['Age-Corrected Standard Score'],
            'toolbox_cogearlycomp_tscore': cogearlycomp['Fully-Corrected T-score'],
        })

    if cogearlycomp:
        data.update({
            'toolbox_cogtotalcomp_uncstd': cogtotalcomp['Uncorrected Standard Score'],
            'toolbox_cogtotalcomp_agestd': cogtotalcomp['Age-Corrected Standard Score'],
            'toolbox_cogtotalcomp_tscore': cogtotalcomp['Fully-Corrected T-score'],
        })

    return data


def _load(project, record_id, event_id, data):



    data[project.def_field] = record_id
    data['redcap_event_name'] = event_id

    try:
        response = project.import_records([data])
        assert 'count' in response
        logger.debug('uploaded')
    except AssertionError as e:
        logger.error('error uploading', record_id, e)


def _extract_regdata(filename):
    data = {}

    try:
        df = pd.read_csv(filename)
    except Exception:
        df = pd.read_excel(filename)

    # Get data from last row
    data = df.iloc[-1].to_dict()

    return data


def _extract_cogscores(filename):
    data = {}

    # Load csv
    try:
        df = pd.read_csv(filename)
    except Exception:
        df = pd.read_excel(filename)

    # Drop instrument duplicates, keeping the last only
    df = df.drop_duplicates(subset='Inst', keep='last')

    # convert to dict of dicts indexed by Instrument
    df = df.dropna(subset=['Inst'])
    df = df.set_index('Inst')
    df = df.fillna('')
    data = df.to_dict('index')

    return data



import logging
from datetime import datetime

import pandas as pd


logger = logging.getLogger('garjus.automations.etl_lifedata.dailyweekly')


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        logger.error(err)
        return False


# Load Life tests and calculate summary measures
def extract_life(records):
    data = {}

    # Load into dataframe
    df = pd.DataFrame(records)

    # Set day using first real notification as baseline, skip 0th
    df['lifedata_notification_time'] = pd.to_datetime(
        df['lifedata_notification_time'], errors='coerce')
    df = df.sort_values('lifedata_notification_time')
    if len(df) > 1:
        df['day'] = (df['lifedata_notification_time'] - df.iloc[1]['lifedata_notification_time']).dt.days
    else:
        df['day'] = (df['lifedata_notification_time'] - df.iloc[0]['lifedata_notification_time']).dt.days

    df['day'] = df['day'] + 1

    # Count complete sessions for the week
    data['life_wlifesesscomp'] = len(df[df['lifedata_responded'] == '1'])

    # Convert format to float so we can math
    df['lifedata_deptot'] = pd.to_numeric(
        df['lifedata_deptot'], errors='coerce')
    df['lifedata_fatiguetot'] = pd.to_numeric(
        df['lifedata_fatiguetot'], errors='coerce')
    df['lifedata_negtot'] = pd.to_numeric(
        df['lifedata_negtot'], errors='coerce')
    df['lifedata_postot'] = pd.to_numeric(
        df['lifedata_postot'], errors='coerce')
    df['lifedata_rumtot'] = pd.to_numeric(
        df['lifedata_rumtot'], errors='coerce')
    df['lifedata_stress'] = pd.to_numeric(
        df['lifedata_stress'], errors='coerce')

    # Calculate range for whole week
    data[f'life_wrange_deptot'] = \
        df.lifedata_deptot.max() - df.lifedata_deptot.min()
    data[f'life_wrange_fatiguetot'] = \
        df.lifedata_fatiguetot.max() - df.lifedata_fatiguetot.min()
    data[f'life_wrange_pnnegtot'] = \
        df.lifedata_negtot.max() - df.lifedata_negtot.min()
    data[f'life_wrange_pnpostot'] = \
        df.lifedata_postot.max() - df.lifedata_postot.min()
    data[f'life_wrange_rumtot'] = \
        df.lifedata_rumtot.max() - df.lifedata_rumtot.min()
    data[f'life_wrange_stresssev'] = \
        df.lifedata_stress.max() - df.lifedata_stress.min()

    # Calculate means for whole week
    data['life_wmean_deptot'] = df.lifedata_deptot.mean()
    data['life_wmean_fatiguetot'] = df.lifedata_fatiguetot.mean()
    data['life_wmean_pnnegtot'] = df.lifedata_negtot.mean()
    data['life_wmean_pnpostot'] = df.lifedata_postot.mean()
    data['life_wmean_rumtot'] = df.lifedata_rumtot.mean()
    data['life_wmean_stresssev'] = df.lifedata_stress.mean()

    # Calculate SD for whole week
    data['life_wsd_deptot'] = df.lifedata_deptot.std()
    data['life_wsd_fatiguetot'] = df.lifedata_fatiguetot.std()
    data['life_wsd_pnnegtot'] = df.lifedata_negtot.std()
    data['life_wsd_pnpostot'] = df.lifedata_postot.std()
    data['life_wsd_rumtot'] = df.lifedata_rumtot.std()
    data['life_wsd_stresssev'] = df.lifedata_stress.std()

    # Calculate CV for whole week
    data[f'life_wcov_deptot'] = \
        df.lifedata_deptot.std() / \
        df.lifedata_deptot.mean()
    data[f'life_wcov_fatiguetot'] = \
        df.lifedata_fatiguetot.std() / \
        df.lifedata_fatiguetot.mean()
    data[f'life_wcov_pnnegtot'] = \
        df.lifedata_negtot.std() / \
        df.lifedata_negtot.mean()
    data[f'life_wcov_pnpostot'] = \
        df.lifedata_postot.std() / \
        df.lifedata_postot.mean()
    data[f'life_wcov_rumtot'] = \
        df.lifedata_rumtot.std() / \
        df.lifedata_rumtot.mean()
    data[f'life_wcov_stresssev'] = \
        df.lifedata_stress.std() / \
        df.lifedata_stress.mean()

    # Loop days 1-7, calculate for each day
    for d in range(1, 8):
        dfd = df[df['day'] == d]

        # Count complete sessions
        data[f'life_d{d}lifesesscomp'] = \
            len(dfd[dfd['lifedata_responded'] == '1'])

        # Calculate ranges
        data[f'life_d{d}range_deptot'] = dfd.lifedata_deptot.max() - dfd.lifedata_deptot.min()
        data[f'life_d{d}range_fatiguetot'] = dfd.lifedata_fatiguetot.max() - dfd.lifedata_fatiguetot.min()
        data[f'life_d{d}range_pnnegtot'] = dfd.lifedata_negtot.max() - dfd.lifedata_negtot.min()
        data[f'life_d{d}range_pnpostot'] = dfd.lifedata_postot.max() - dfd.lifedata_postot.min()
        data[f'life_d{d}range_rumtot'] = dfd.lifedata_rumtot.max() - dfd.lifedata_rumtot.min()
        data[f'life_d{d}range_stresssev'] = dfd.lifedata_stress.max() - dfd.lifedata_stress.min()

        # Calculate means
        data[f'life_d{d}mean_deptot'] = dfd.lifedata_deptot.mean()
        data[f'life_d{d}mean_fatiguetot'] = dfd.lifedata_fatiguetot.mean()
        data[f'life_d{d}mean_pnnegtot'] = dfd.lifedata_negtot.mean()
        data[f'life_d{d}mean_pnpostot'] = dfd.lifedata_postot.mean()
        data[f'life_d{d}mean_rumtot'] = dfd.lifedata_rumtot.mean()
        data[f'life_d{d}mean_stresssev'] = dfd.lifedata_stress.mean()

        # Calculate SD
        data[f'life_d{d}sd_deptot'] = dfd.lifedata_deptot.std()
        data[f'life_d{d}sd_fatiguetot'] = dfd.lifedata_fatiguetot.std()
        data[f'life_d{d}sd_pnnegtot'] = dfd.lifedata_negtot.std()
        data[f'life_d{d}sd_pnpostot'] = dfd.lifedata_postot.std()
        data[f'life_d{d}sd_rumtot'] = dfd.lifedata_rumtot.std()
        data[f'life_d{d}sd_stresssev'] = dfd.lifedata_stress.std()

        # Calculate CV
        data[f'life_d{d}cov_deptot'] =dfd.lifedata_deptot.std() / dfd.lifedata_deptot.mean()
        data[f'life_d{d}cov_fatiguetot'] = dfd.lifedata_fatiguetot.std() / dfd.lifedata_fatiguetot.mean()
        data[f'life_d{d}cov_pnnegtot'] = dfd.lifedata_negtot.std() / dfd.lifedata_negtot.mean()
        data[f'life_d{d}cov_pnpostot'] =  dfd.lifedata_postot.std() / dfd.lifedata_postot.mean()
        data[f'life_d{d}cov_rumtot'] = dfd.lifedata_rumtot.std() / dfd.lifedata_rumtot.mean()
        data[f'life_d{d}cov_stresssev'] = dfd.lifedata_stress.std() / dfd.lifedata_stress.mean()

    # Return etl data as strings
    for k, v in data.items():
        data[k] = str(v)

    return data


def process_project(project):
    results = []
    def_field = project.def_field
    fields = [
        def_field,
        'lifedata_responded',
        'life_wlifesesscomp'
    ]

    subj2id = {}
    subjects = []

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        subj2id = {x[sec_field]: x[def_field] for x in rec if x[sec_field]}
        subjects = list(set([x[sec_field] for x in rec if x[sec_field]]))
        subjects = sorted(subjects)
    else:
        rec = project.export_records(fields=[def_field])
        subj2id = {x[def_field]: x[def_field] for x in rec if x[def_field]}
        subjects = sorted(list(set([x[def_field] for x in rec])))

    # Get records
    all_records = project.export_records(fields=fields, forms=['ema_lifedata_survey'])

    # Get the repeating records
    life_records = [x for x in all_records if x['redcap_repeat_instance'] and x['redcap_repeat_instrument'] == 'ema_lifedata_survey']

    # Process each subject
    for subj in subjects:
        subj_id = subj2id[subj]
        subj_events = sorted(list(set([x['redcap_event_name'] for x in all_records if x[def_field] == subj_id])))
        subj_life = [x for x in life_records if x[def_field] == subj_id]

        # Iterate subject events
        for event_id in subj_events:

            # Find existing
            found = [x for x in all_records if (x[def_field] == subj_id) and (x['redcap_event_name'] == event_id) and (x.get('life_wlifesesscomp', False))]
            if len(found) > 0:
                logger.debug(f'found existing:{subj}:{event_id}')
                continue

            # Get repeat records
            repeats = [x for x in subj_life if (x['redcap_event_name'] == event_id)]
            if len(repeats) == 0:
                # no repeats to count
                logger.debug(f'no repeats found:{subj}:{event_id}')
                continue

            finished = [x for x in subj_life if (x['redcap_event_name'] == event_id) and (x.get('lifedata_responded', '0') == '1')]
            count_finished = len(finished)

            logger.debug(f'{subj}:{event_id}:{count_finished=}')

            data = {
                def_field: subj_id,
                'redcap_event_name': event_id,
                'life_dailyweekly_complete': '2',
            }

            if count_finished > 0:
                # Calculate daily and weekly
                data.update(extract_life(finished))
            else:
                logger.debug(f'no finished records found:{subj}:{event_id}')
                data.update({'life_wlifesesscomp': '0'})

            result = _load(project, [data])
            if result is True:
                results.append({
                    'result': 'COMPLETE',
                    'description': 'life_dailyweekly',
                    'category': 'life_dailyweekly',
                    'subject': subj_id,
                    'event': event_id
                })

    return results



import io
import logging
import os, sys
import tempfile

import numpy as np
import pandas as pd
import redcap


logger = logging.getLogger('garjus.automations.ACOBA_posner')


VALID = 'Valid'
INVALID = 'Invalid'
NOCUE = 'NoCue'
NEUTRAL = 'Neutral'

FORM_INCOMPLETE = 0
FORM_UNVERIFIED = 1
FORM_COMPLETE = 2

API_URL = 'https://redcap.vanderbilt.edu/api/'
API_KEY = sys.argv[1]


REDCAP_EVENTS = [
    'screening_arm_1',
    'baseline_arm_1',
]

REDCAP_FIELDS = [
    'participant_id',
    'posner_tabfile',
    'posner_complete',
    'posner_alerting_rt',
]

def read_edat(edat_path):
    _encoding = 'utf-8'
    _skiprows = 0
    first_field = 'ExperimentName'
    
    logger.info(f'edat_path={edat_path}')
    
    # Determine how many rows to skip prior to header
    with io.open(edat_path, encoding=_encoding) as _f:
        for line in _f:
            if line.startswith(first_field):
                break
            else:
                _skiprows += 1
                
    # Load Data
    df = pd.read_table(edat_path, sep='\t', encoding=_encoding, skiprows=_skiprows, header=0)
    return df


def load_edat(edat_path):
    logger.debug(f'parsing edat:{edat_path}')
    df = read_edat(edat_path)

    # Fix column names
    df.columns = df.columns.map(lambda x: x.replace('.', '_').replace(']', '_').replace('[', '_'))

    return df


def apply_columns(row):
    type_col = 'Code'
    row_type = row[type_col]

    row['onset'] = row['Target_OnsetTime']
    row['rt'] = row['Target_RT']
    row['acc'] = row['Target_ACC']

    if row_type == 'valid':
        row['trial_type'] = VALID
    elif row_type == 'invalid':
        row['trial_type'] = INVALID
    elif row_type == 'nocue':
        row['trial_type'] = NOCUE
    elif row_type == 'neutral':
        row['trial_type'] = NEUTRAL
    else:
        raise ValueError('unknown trial type')

    row['duration'] = 0

    return row


def download_file(proj, record_id, event_id, field_id, tmpdir):
    filename = os.path.join(tmpdir, record_id+'_'+event_id+'_'+field_id)
    
    logger.debug(f'downloading file:{record_id}:{event_id}:{field_id}')
    
    try:
        cont,hdr = proj.export_file(record=record_id, event=event_id, field=field_id)
        if cont == '':
            raise RedcapError
    except RedcapError:
        logger.error('downloading file')
        return None
    else:
        with open(filename, 'wb') as f:
            f.write(cont)

        return filename


def parse_posner_edat(edat_file):
    df = load_edat(edat_file)
    df = df.apply(apply_columns, axis=1)
    start_offset = df.iloc[0].Target_OnsetTime
    df['onset'] = (df['onset'] - start_offset) / 1000.0
    df = df[['trial_type', 'onset', 'duration', 'rt', 'acc']]

    # Drop Fixation rows?
    #df = df[df['trial_type'].isin(['Valid','Invalid','NoCue','Neutral'])]
    
    return df


def extract_posner(proj, record_id, event_id, file_field, tmpdir):
    data = {}

    tab_file = download_file(proj, record_id, event_id, file_field, tmpdir)

    if tab_file is None:
        return None

    df = parse_posner_edat(tab_file)

    # Get subset of only accurate trials
    df1 = df[(df.acc == 1)]

    for condition in [VALID, INVALID, NOCUE, NEUTRAL]:
        data[condition + '_ACC'] = df[(df['trial_type'] == condition)].acc.mean()
        if data[condition + '_ACC'] > 0:
            # Get median from accurate trials
            data[condition + '_RT_median'] = df1[(df1['trial_type'] == condition)].rt.median()
        else:
            data[condition + '_RT_median'] = np.nan

    return data


def load_posner(proj, record_id, event_id, file_field, tmpdir):

    posner_data = extract_posner(proj, record_id, event_id, file_field, tmpdir)

    if posner_data is None:
        logger.error('posner extract failed')
        return

    # Import the data
    data = {}
    data[proj.def_field] = record_id
    data['redcap_event_name'] = event_id
    data['posner_complete'] = FORM_COMPLETE    
    for condition in [VALID, INVALID, NOCUE, NEUTRAL]:
        data['posner_' + condition.lower() + '_rt_median'] = posner_data[condition + '_RT_median']
        data['posner_' + condition.lower() + '_acc'] = posner_data[condition + '_ACC']

    # Append additional RT measures
    data['posner_alerting_rt'] = data['posner_neutral_rt_median'] - data['posner_nocue_rt_median']
    data['posner_reorienting_rt'] = data['posner_valid_rt_median'] - data['posner_invalid_rt_median']

    try:
        response = proj.import_records([data])
        assert 'count' in response
        logger.info(f'uploaded:{record_id}:{event_id}')
    except AssertionError as err:
        logger.error(f'uploaded:{record_id}:{event_id}')


logging.basicConfig(
    format='%(asctime)s - %(levelname)s:%(name)s:%(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')

tmpdir = tempfile.mkdtemp()
logger.debug(f'tmpdir:{tmpdir}')

# Load the REDCap database
logger.info('INFO:loading REDCap...')
proj = redcap.Project(API_URL, API_KEY)
data = proj.export_records(events=REDCAP_EVENTS, fields=REDCAP_FIELDS)

for r in data:
    if not r['posner_tabfile']:
        logger.debug(f'no file:{r}')
        continue

    if not r['posner_tabfile'].endswith('.txt'):
        logger.debug(f'wrong format:{r}')
        continue

    if r['posner_tabfile'] == 'CONVERT_FAILED.txt':
        logger.debug(f'found CONVERT_FAILED')
        continue

    if r['posner_tabfile'] == 'MISSING_DATA.txt':
        logger.debug(f'found MISSING_DATA')
        continue

    if r['posner_alerting_rt']:
        logger.debug(f'already extracted:{r[proj.def_field]}:{r["posner_tabfile"]}')
        continue


    logger.info(f'loading:{r["posner_tabfile"]}')
    load_posner(proj, r[proj.def_field], r['redcap_event_name'], 'posner_tabfile', tmpdir)
       
logger.info('DONE!')



'''Subjects from REDCap.'''
import logging

import pandas as pd
import numpy as np

from . import covariates


logger = logging.getLogger('garjus.subjects')

IDENTIFIER_FIELDS = [
        'record_id',
        'care_subject_number',
        'acoba_subject_number',
        'r21_subject_number',
        'depressedmind_subject_number',
        'rem_subject_number',
        'depmind2_subject_number',
        'depmind3_subject_number',
        'd3_subject_number',
        'guid'
    ]


def load_fallypride_data(rc):
    return covariates._load_fallypride(rc)


def load_gait_data(rc):
    def_field = rc.def_field
    fields = [
        def_field,
        'gaitrite_comments',
        'gaitrite_velocity',
    ]

    events = [
        'baseline_arm_1',
        'baseline_arm_2',
        'baseline_arm_3',
        'baseline_arm_4'
    ]

    rec = rc.export_records(fields=fields)

    # Only gaitrite records
    rec = [x for x in rec if x['redcap_repeat_instrument'] == 'gaitrite']

    # Filter by selected events
    rec = [x for x in rec if x['redcap_event_name'] in events]

    rec = [x for x in rec if \
        x['gaitrite_comments'].lower().startswith('standard') or \
        x['gaitrite_comments'].lower().startswith('first')
    ]

    dfg = pd.DataFrame(rec, columns=fields)

    dfg = dfg.astype(str)
    
    dfg = dfg.drop_duplicates(subset=[def_field], keep='first')

    dfg['SPEED'] = dfg['gaitrite_velocity']

    return dfg[[def_field, 'SPEED']]


def load_R21Perfusion(garjus):
    R21_SCREEN_FIELDS = [
        'participant_id',
        'study_ident',
        'madrs_same',
        'ma_tot',
        'crit_meets',
        'critcon_meets']

    R21_BASE_FIELDS = [
        'participant_id',
        'madrs_same',
        'ma_tot',
        'age',
        'dob',
        'sex_xcount']

    rc = garjus.primary('R21Perfusion')

    screenR = rc.export_records(
        raw_or_label='label',
        format_type='df',
        fields=R21_SCREEN_FIELDS,
        events=['screen_arm_1'])
    baseR = rc.export_records(
        raw_or_label='label',
        format_type='df',
        fields=R21_BASE_FIELDS,
        events=['baseline_arm_1'])

    # Change index
    screenR.reset_index(inplace=True)
    screenR.set_index('participant_id', inplace=True)
    screenR.drop(['redcap_event_name'], inplace=True, axis=1)

    # Change index
    baseR.reset_index(inplace=True)
    baseR.set_index('participant_id', inplace=True)
    baseR.drop(['redcap_event_name'], inplace=True, axis=1)

    # Set the group
    screenR = screenR.apply(get_group_R21Perfusion, axis=1)

    # Prep to later merge MADRS
    screenR['ma_tot_screen'] = screenR['ma_tot']
    screenR['madrs_same_screen'] = screenR['madrs_same']
    screenR.drop(['ma_tot','madrs_same'], inplace=True, axis=1)
    baseR['ma_tot_base'] = baseR['ma_tot']
    baseR['madrs_same_base'] = baseR['madrs_same']
    baseR.drop(['ma_tot','madrs_same'], inplace=True, axis=1)

    # Merge Baseline and Screen
    dataR = pd.merge(
        screenR,
        baseR,
        how='outer',
        left_index=True,
        right_index=True,
        sort=True)

    # Merge MADRS
    dataR = dataR.apply(get_base_madrs, axis=1)
    dataR.drop(
        ['ma_tot_base', 'madrs_same_base', 'ma_tot_screen', 'madrs_same_screen'],
        inplace=True, axis=1)

    # Change index to ID
    dataR['study_ident'] = dataR['study_ident'].astype(int).astype(str)
    dataR.rename(columns={'study_ident': 'ID'}, inplace=True)
    dataR.reset_index(inplace=True)
    dataR.set_index('ID', inplace=True)
    dataR.drop(['participant_id'], inplace=True, axis=1)

    # Set common fields
    dataR['PROJECT'] = 'R21Perfusion'
    dataR['AGE'] = dataR['age']
    dataR = dataR[~dataR.dob.isna()]
    dataR['DOB'] = pd.to_datetime(dataR['dob'])
    dataR['SEX'] = dataR['sex_xcount'].map({'Male': 'M', 'Female': 'F'})

    return dataR


def get_group_R21Perfusion(row):
    if row['crit_meets'] == 'Yes' and row['critcon_meets'] == 'Yes':
        print('ERROR:cannot be both DEPRESSED and CONTROL')
    elif row['crit_meets'] == 'Yes':
        row['GROUP'] = 'Depress'
    elif row['critcon_meets'] == 'Yes':
        row['GROUP'] = 'Control'

    return row


def get_base_madrs(row):    
    base_madrs = row['ma_tot_base']
    screen_madrs = row['ma_tot_screen']
    screen_same = row['madrs_same_screen']

    if np.isfinite(base_madrs):
        row['ma_tot'] = base_madrs
    elif ~np.isfinite(base_madrs) and np.isfinite(screen_madrs) and screen_same:
        row['ma_tot'] = screen_madrs
    elif ~np.isfinite(base_madrs) and np.isfinite(screen_madrs):
        print('using screen madrs:'+str(row['study_ident']))
        row['ma_tot'] = screen_madrs
    else:
        pass

    return row


def load_CAARE(garjus):
    C1_SCREEN_FIELDS = [
        'participant_id',
        'study_ident',
        'madrs_same',
        'ma_tot',
        'mmse_total',
        'age',
        'dob',
        'sex_xcount']

    C1_BASE_FIELDS = [
        'participant_id',
        'madrs_same',
        'ma_tot']

    C2_SCREEN_FIELDS = [
        'participant_id',
        'study_ident',
        'madrs_same',
        'ma_tot',
        'mmse_total',
        'age',
        'dob',
        'sex_xcount']

    C2_BASE_FIELDS = [
        'participant_id',
        'madrs_same',
        'ma_tot']

    rc = garjus.primary('TAYLOR_CAARE')

    screenC1 = rc.export_records(
        raw_or_label='label', format_type='df', fields=C1_SCREEN_FIELDS, events=['screen_arm_1'])
    baseC1 = rc.export_records(
        raw_or_label='label', format_type='df', fields=C1_BASE_FIELDS, events=['baseline_arm_1'])
    screenC2 = rc.export_records(
        raw_or_label='label', format_type='df', fields=C2_SCREEN_FIELDS, events=['screen_arm_2'])
    baseC2 = rc.export_records(
        raw_or_label='label', format_type='df', fields=C2_BASE_FIELDS, events=['baseline_arm_2'])

    # Change index
    screenC1.reset_index(inplace=True)
    screenC1.set_index('participant_id', inplace=True)
    screenC1.drop(['redcap_event_name'], inplace=True, axis=1)

    # Change index
    baseC1.reset_index(inplace=True)
    baseC1.set_index('participant_id', inplace=True)
    baseC1.drop(['redcap_event_name'], inplace=True, axis=1)

    # Deal with MADRS
    screenC1['ma_tot_screen'] = screenC1['ma_tot']
    screenC1['madrs_same_screen'] = screenC1['madrs_same']
    screenC1.drop(['ma_tot','madrs_same'], inplace=True, axis=1)
    baseC1['ma_tot_base'] = baseC1['ma_tot']
    baseC1['madrs_same_base'] = baseC1['madrs_same']
    baseC1.drop(['ma_tot','madrs_same'], inplace=True, axis=1)

    # Now merge screen and base
    dataC1 = pd.merge(screenC1, baseC1, how='outer', left_index=True, right_index=True, sort=True)
    dataC1 = dataC1[np.isfinite(dataC1['study_ident'])]
    dataC1['study_ident'] = dataC1['study_ident'].astype(int).astype(str)
    dataC1 = dataC1.apply(get_base_madrs, axis=1)

    # Change index
    screenC2.reset_index(inplace=True)
    screenC2.set_index('participant_id', inplace=True)
    screenC2.drop(['redcap_event_name'], inplace=True, axis=1)

    # Change index
    baseC2.reset_index(inplace=True)
    baseC2.set_index('participant_id', inplace=True)
    baseC2.drop(['redcap_event_name'], inplace=True, axis=1)

    # Deal with MADRS
    screenC2['ma_tot_screen'] = screenC2['ma_tot']
    screenC2['madrs_same_screen'] = screenC2['madrs_same']
    screenC2.drop(['ma_tot','madrs_same'], inplace=True, axis=1)
    baseC2['ma_tot_base'] = baseC2['ma_tot']
    baseC2['madrs_same_base'] = baseC2['madrs_same']
    baseC2.drop(['ma_tot','madrs_same'], inplace=True, axis=1)

    # Now merge screen and base
    dataC2 = pd.merge(
        screenC2, baseC2,
        how='outer', left_index=True, right_index=True, sort=True)
    dataC2['study_ident'] = dataC2['study_ident'].astype(int).astype(str)
    dataC2 = dataC2.apply(get_base_madrs, axis=1)

    # Concat arms
    dataC1['GROUP'] = 'Depress'
    dataC2['GROUP'] = 'Control'
    dataC = pd.concat([dataC1, dataC2], sort=True)

    # Change index to ID
    dataC.rename(columns={'study_ident': 'ID'}, inplace=True)
    dataC.reset_index(inplace=True)
    dataC.set_index('ID', inplace=True)
    dataC.drop(['participant_id'], inplace=True, axis=1)

    # Drop unused columns
    dataC.drop(
        ['ma_tot_base', 'madrs_same_base', 'ma_tot_screen', 'madrs_same_screen'],
        inplace=True, axis=1)

    dataC['PROJECT'] = 'TAYLOR_CAARE'
    dataC['AGE'] = dataC['age']
    dataC['DOB'] = pd.to_datetime(dataC['dob'])
    dataC['SEX'] = dataC['sex_xcount'].map({'Male': 'M', 'Female': 'F'})

    dataC = dataC[~dataC.DOB.isna()]

    return dataC


def load_MDDHx():
    controls = [
        '2503', '2505', '2510', '2511', '2513', '2518', '2520', '2522', '2526',
        '2527', '2529', '2531', '2532', '2535', '2536', '2537', '2538', '2539',
        '2540', '2541', '2542', '2543', '2544', '2547', '2551', '2552', '2553',
        '2554', '2556', '2558', '2561', '2563', '2569', '2574', '2575', '2577',
        '2578', '2579', '2580', '2581', '2582', '2585', '2586', '2588', '2592'
        ]

    age = [
        64, 62, 63, 69, 68, 66, 74, 54, 60, 68, 70, 72, 57, 54, 63, 58, 53, 53,
        61, 67, 55, 54, 56, 57, 66, 65, 58, 58, 53, 71, 56, 54, 59, 55, 51, 67,
        52, 70, 60, 59, 56, 66, 67, 70, 58]

    df = pd.DataFrame({'ID': controls})
    df['GROUP'] = 'Control'
    df['PROJECT'] = 'NewhouseMDDHx'
    df['AGE'] = age
    df['SEX'] = 'F'

    # Finish up
    df = df.sort_values('ID')
    df = df.drop_duplicates()
    df = df.set_index('ID')

    return df


def load_COGD(garjus):
    def_field = 'record_id'
    dob_field = 'backg_dob'
    sex_field = 'backg_sex'
    date_field = 'mri_date'

    project_redcap = garjus.primary('COGD')

    _fields = [dob_field, sex_field, date_field]
    rec = project_redcap.export_records(fields=_fields, raw_or_label='label')

    try:
        rec = [x for x in rec if x[dob_field]]
    except KeyError as err:
        logger.debug(f'cannot access dob:{dob_field}:{err}')

    _fields = [def_field, dob_field, sex_field, 'redcap_event_name',  'backg_sex___1', 'backg_sex___2']
    df = pd.DataFrame(rec, columns=_fields)
  
    df['SEX'] = 'UNKNOWN'
    df.loc[df['backg_sex___1'] == 'Checked', 'SEX'] = 'M'
    df.loc[df['backg_sex___2'] == 'Checked' , 'SEX'] = 'F'

    # all are depressed
    df['GROUP'] = 'Depress'
    df['PROJECT'] = 'COGD'
    df['ID'] = df[def_field]

    # Load MRI records to get first date
    fields = [def_field, date_field]
    rec = project_redcap.export_records(fields=fields, raw_or_label='label')
    rec = [x for x in rec if x[date_field]]
    dfm = pd.DataFrame(rec, columns=fields)
    dfm = dfm.astype(str)
    dfm = dfm.sort_values(date_field)
    dfm = dfm.drop_duplicates(subset=[def_field], keep='first')

    # Merge in date
    df = pd.merge(df, dfm, how='left', on=def_field)

    # Drop intermediate columns
    drop_columns = [def_field, 'redcap_event_name', 'backg_sex___1', 'backg_sex___2', sex_field]
    drop_columns = [x for x in drop_columns if x and x in df.columns]
    df = df.drop(columns=drop_columns)

    # Exclude incomplete data
    df = df.dropna()

    # Calculate age at baseline
    df[dob_field] = pd.to_datetime(df[dob_field])
    df[date_field] = pd.to_datetime(df[date_field])
    df['AGE'] = (
        df[date_field] - df[dob_field]
    ).values.astype('<m8[Y]').astype('int').astype('str')

    df['DOB'] = df[dob_field]

    return df


def load_standard(garjus, project, include_dob=False):
    project_redcap = garjus.primary(project)

    if not project_redcap:
        logger.debug(f'project redcap not found:{project}')
        return pd.DataFrame([], columns=['ID', 'PROJECT', 'GROUP'])

    def_field = project_redcap.def_field
    sec_field = project_redcap.export_project_info()['secondary_unique_field']
    guid_field = None
    sex_field = None
    dob_field = None
    date_field = None
    field_names = project_redcap.field_names

    if 'guid' in field_names:
        guid_field = 'guid'
    elif 'guid_5b4861' in field_names:
        guid_field = 'guid_5b4861'

    if 'dob' in field_names:
        dob_field = 'dob'
    elif 'dob_sub' in field_names:
        dob_field = 'dob_sub'
    elif 'backg_dob' in field_names:
        dob_field = 'backg_dob'

    if 'sex_xcount' in field_names:
        sex_field = 'sex_xcount'
    elif 'dems_sex' in field_names:
        sex_field = 'dems_sex'
    elif 'sex_demo' in field_names:
        sex_field = 'sex_demo'

    if 'mri_date' in field_names:
        date_field = 'mri_date'

    # Load subject records from redcap
    fields = [def_field]

    if sec_field:
        fields.append(sec_field)

    if guid_field:
        fields.append(guid_field)

    if dob_field:
        fields.append(dob_field)

    if sex_field:
        fields.append(sex_field)

    rec = project_redcap.export_records(fields=fields, raw_or_label='label')

    # Ignore records without secondary ID
    if sec_field:
        rec = [x for x in rec if x[sec_field]]

    # Make data frame
    if project_redcap.is_longitudinal:
        df = pd.DataFrame(rec, columns=fields + ['redcap_event_name'])
    else:
        df = pd.DataFrame(rec, columns=fields)

    # Set the project
    df['PROJECT'] = project

    # Determine group
    df['GROUP'] = 'UNKNOWN'

    if project in ['DepMIND2', 'DepMIND3']:
        # All are depressed
        df['GROUP'] = 'Depress'

        # Subset of events where demographics are found
        df = df[df.redcap_event_name.isin([
            'Screening (Arm 1: Blinded Phase)',
        ])]
    elif project == 'D3':
        # Load gait velocity
        dfg = load_gait_data(project_redcap)

        # Merge in gait data
        df = pd.merge(df, dfg, how='left', on=def_field)

        # Load fallypride roi data
        dfr = load_fallypride_data(project_redcap)

        # Use arm/events names to determine group
        df['GROUP'] = df['redcap_event_name'].map({
            'Screening (Arm 2: VUMC Never Depressed)': 'Control',
            'Screening (Arm 1: VUMC Currently Depressed)': 'Depress',
            'Screening (Arm 4: UPMC Never Depressed)': 'Control',
            'Screening (Arm 3: UPMC Currently Depressed)': 'Depress',
        })

        df = df[df.GROUP.isin(['Depress', 'Control'])]

        # Merge in fallypride data
        df = pd.merge(df, dfr, how='left', left_on=sec_field, right_on='ID')
        print(df)
    elif project == 'REMBRANDT':
        # Use arm/events names to determine which arm
        df['GROUP'] = df['redcap_event_name'].map({
            'Screening (Arm 3: Longitudinal Phase: Remitted)': 'Depress',
            'Screening (Arm 2: Longitudinal Phase: Never Depressed)': 'Control',
        })
        df = df[df.GROUP.isin(['Depress', 'Control'])]
    elif project == 'CHAMP':
        # Subset of events where demographics are found
        df = df[df.redcap_event_name.isin([
            'Screening (Arm 1: Screening)',
        ])]

    # Load MRI records to get first date
    if dob_field and date_field:
        fields = [def_field, date_field]
        rec = project_redcap.export_records(fields=fields, raw_or_label='label')
        rec = [x for x in rec if x[date_field]]
        dfm = pd.DataFrame(rec, columns=fields)
        dfm = dfm.astype(str)
        dfm = dfm.sort_values(date_field)
        dfm = dfm.drop_duplicates(subset=[def_field], keep='first')

        # Merge in date
        df = pd.merge(df, dfm, how='left', on=def_field)

        df[sex_field] = df[sex_field].fillna('UNKNOWN')

        # Calculate age at baseline
        df[dob_field] = pd.to_datetime(df[dob_field])
        df[date_field] = pd.to_datetime(df[date_field])
        df['AGE'] = (
            df[date_field] - df[dob_field]
        ).values.astype('<m8[Y]').astype('int').astype('str')

        # Replace sentinel value with blank
        df.loc[df.AGE.astype('int') < 0, 'AGE'] = ''

        if include_dob:
            df['DOB'] = df[dob_field]

    if sex_field:
        df['SEX'] = df[sex_field].map({
            'Male': 'M',
            'Female': 'F',
            'UNKNOWN': 'U'
        })

    if guid_field:
        df['GUID'] = df[guid_field]

    if sec_field:
        df['ID'] = df[sec_field]
    else:
        df['ID'] = df[def_field]

    # Drop intermediate columns
    drop_columns = [def_field, 'redcap_event_name', dob_field, date_field, sec_field, guid_field, sex_field]
    drop_columns = [x for x in drop_columns if x and x in df.columns]
    df = df.drop(columns=drop_columns)

    # Finish up
    df = df.sort_values('ID')
    df = df.drop_duplicates()
    return df


def load_subjects(garjus, project, include_dob=False):
    if project == 'COGD':
        df = load_COGD(garjus)
    elif project == 'NewhouseMDDHx':
        df = load_MDDHx()
    elif project == 'R21Perfusion':
        df = load_R21Perfusion(garjus)        
    elif project == 'TAYLOR_CAARE':
        df = load_CAARE(garjus)
    else:
        df = load_standard(garjus, project, include_dob)

    try:
        # Get identifier database record id
        logger.debug(f'getting identifier database:{project}')
        rc = garjus.identifier_database()
        logger.debug(f'loading records from identifier database:{project}')
        rec = rc.export_records(fields=IDENTIFIER_FIELDS)
        logger.debug(f'handling records from identifier database:{project}')
        dfi = pd.DataFrame(rec)

        # Set the ID data to match based on project
        if project == 'REMBRANDT':
            dfi['ID'] = dfi['rem_subject_number']
        elif project == 'TAYLOR_CAARE':
            dfi['ID'] = dfi['care_subject_number']
        elif project == 'ACOBA':
            dfi['ID'] = dfi['acoba_subject_number']
        elif project == 'R21Perfusion':
            dfi['ID'] = dfi['r21_subject_number']
        elif project == 'DepMIND':
            dfi['ID'] = dfi['depressedmind_subject_number']
        elif project == 'DepMIND2':
            dfi['ID'] = dfi['depmind2_subject_number']
        elif project == 'DepMIND3':
            dfi['ID'] = dfi['depmind3_subject_number']
        elif project == 'D3':
            dfi['ID'] = dfi['d3_subject_number']
        else:
            raise Exception('could not match project to identifer database column')

        # Set the identifier id as redcap record id
        dfi['identifier_id'] = dfi['record_id']

        # Get just the identifiers without duplicates
        dfi = dfi[['identifier_id', 'ID']].drop_duplicates(subset=['ID'], keep='first')

        # Merge the subjects with identifier records
        df = pd.merge(df, dfi, how='left', on='ID', validate="1:1")
    except Exception as err:
        logger.debug(f'failed to load identifier database:{project}:{err}')

    return df



import logging

import pandas as pd
import plotly
import plotly.graph_objs as go
import plotly.subplots
from dash import Input, Output, callback, dcc, html, dash_table as dt
import dash_bootstrap_components as dbc

from .. import utils
from ..shared import STATUS2RGB
from . import data


logger = logging.getLogger('dashboard.queue')

# STATUS2EMO???


STATUSES = [
    'FAILED',
    'COMPLETE',
    'COMPLETED',
    'UPLOADING',
    'RUNNING',
    'PENDING',
    'WAITING',
    'QUEUED',
    'UNKNOWN',
]


def get_graph_content(df):
    status2rgb = {k: STATUS2RGB[k] for k in STATUSES}

    # Make a 1x1 figure
    fig = plotly.subplots.make_subplots(rows=1, cols=1)

    dfp = pd.pivot_table(
        df,
        index='PROCTYPE',
        values='LABEL',
        columns=['STATUS'],
        aggfunc='count',
        fill_value=0)

    for status, color in status2rgb.items():
        ydata = sorted(dfp.index)
        if status not in dfp:
            xdata = [0] * len(dfp.index)
        else:
            xdata = dfp[status]

        fig.append_trace(
            go.Bar(
                x=xdata,
                y=ydata,
                name='{} ({})'.format(status, sum(xdata)),
                marker=dict(color=color),
                opacity=0.9, orientation='h'),
            1,
            1
        )

    fig['layout'].update(barmode='stack', showlegend=True)

    graph = dcc.Graph(figure=fig)

    return dbc.Spinner(id="loading-queue-graph", children=[graph])


def get_content():
    COLUMNS = [
        'ID',
        'LABEL',
        'STATUS',
        'WALLTIME',
        'MEMREQ',
        'PROJECT',
        'PROCESSOR',
        'USER',
    ]

    columns = utils.make_columns(COLUMNS)

    # Format columns with links as markdown text
    for i, c in enumerate(columns):
        if c['name'] in ['ID']:
            columns[i]['type'] = 'text'
            columns[i]['presentation'] = 'markdown'

    content = [
        dbc.Row([
            dbc.Col(
                dbc.Button(
                    'Refresh Data',
                    id='button-queue-refresh',
                    outline=True,
                    color='primary',
                    size='sm',
                ),
            ),
            dbc.Col(
                dbc.Switch(
                    id='switch-queue-graph',
                    label='Graph',
                    value=False,
                ),
                align='center',
            ),
        ]),
        dbc.Row([
            dbc.Col(
                [
                    dbc.Stack([
                        dcc.Dropdown(
                            id='dropdown-queue-proj',
                            multi=True,
                            placeholder='Select Projects',
                        ),
                        dcc.Dropdown(
                            id='dropdown-queue-proc',
                            multi=True,
                            placeholder='Select Processing Types',
                        ),
                        dcc.Dropdown(
                            id='dropdown-queue-user',
                            multi=True,
                            placeholder='Select Users',
                        ),
                    ]),
                ],
                width=3
            ),
            dbc.Col(html.Div(id='container-queue-graph', children=[])),
        ]),
        dbc.Spinner(id="loading-queue-table", children=[
            dbc.Label('Loading...', id='label-queue-rowcount1'),
        ]),
        dt.DataTable(
            cell_selectable=False,
            columns=columns,
            data=[],
            page_action='none',
            sort_action='native',
            id='datatable-queue',
            style_cell={
                'textAlign': 'center',
                'padding': '5px 5px 0px 5px',
            #    'width': '30px',
            #    'overflow': 'hidden',
            #    'textOverflow': 'ellipsis',
            #    'height': 'auto',
            #    'minWidth': '40',
            #    'maxWidth': '60'
            },
            style_data_conditional=[
                {'if': {'column_id': 'LABEL'}, 'textAlign': 'left'},
            #    {'if': {'filter_query': '{STATUS} = "QUEUED"'},  'backgroundColor': STATUS2HEX['WAITING']},
            #    {'if': {'filter_query': '{STATUS} = "RUNNING"'},  'backgroundColor': STATUS2HEX['RUNNING']},
            #    {'if': {'filter_query': '{STATUS} = "WAITING"'},  'backgroundColor': STATUS2HEX['WAITING']},
            #    {'if': {'filter_query': '{STATUS} = "PENDING"'},  'backgroundColor': STATUS2HEX['PENDING']},
            #    {'if': {'filter_query': '{STATUS} = "UNKNOWN"'},  'backgroundColor': STATUS2HEX['UNKNOWN']},
            #    {'if': {'filter_query': '{STATUS} = "FAILED"'},   'backgroundColor': STATUS2HEX['FAILED']},
            #    {'if': {'filter_query': '{STATUS} = "COMPLETE"'}, 'backgroundColor': STATUS2HEX['COMPLETE']},
            #    {'if': {'column_id': 'STATUS', 'filter_query': '{STATUS} = ""'}, 'backgroundColor': 'white'}
            ],
            style_header={
                'fontWeight': 'bold',
                'padding': '5px 15px 0px 10px',
            },
            export_format='xlsx',
            export_headers='names',
            export_columns='visible',
        ),
        dbc.Label('Get ready...', id='label-queue-rowcount2'),
    ]

    return content


def load_data(refresh=False, hidedone=True):
    return data.load_data(refresh=refresh, hidedone=hidedone)


def load_options(df):
    options = {}

    for k in ['PROCTYPE', 'USER', 'PROJECT']:
        # Get a unique list of strings with blanks removed
        koptions = df[k].unique()
        koptions = [str(x) for x in koptions]
        koptions = [x for x in koptions if x]
        options[k] = sorted(koptions)

    return options


def filter_data(df, selected_proj, selected_proc, selected_user):
    return data.filter_data(
        df, selected_proj, selected_proc, selected_user)


@callback(
    [Output('dropdown-queue-proc', 'options'),
     Output('dropdown-queue-proj', 'options'),
     Output('dropdown-queue-user', 'options'),
     Output('datatable-queue', 'data'),
     Output('container-queue-graph', 'children'),
     Output('label-queue-rowcount1', 'children'),
     Output('label-queue-rowcount2', 'children'),
    ],
    [
     Input('dropdown-queue-proc', 'value'),
     Input('dropdown-queue-proj', 'value'),
     Input('dropdown-queue-user', 'value'),
     Input('switch-queue-graph', 'value'),
     Input('button-queue-refresh', 'n_clicks'),
    ]
)
def update_queue(
    selected_proc,
    selected_proj,
    selected_user,
    selected_graph,
    n_clicks
):
    refresh = False
    graph_content = []

    logger.debug('update_queue')

    # Load data
    if utils.was_triggered('button-queue-refresh'):
        # Refresh data if refresh button clicked
        logger.debug('queue refresh:clicks={}'.format(n_clicks))
        refresh = True

    logger.debug('loading data')
    df = load_data(refresh=refresh)

    # Update lists of possible options for dropdowns (could have changed)
    # make these lists before we filter what to display
    options = load_options(df)
    proj = utils.make_options(options['PROJECT'])
    proc = utils.make_options(options['PROCTYPE'])
    user = utils.make_options(options['USER'])

    # Filter data based on dropdown values
    df = filter_data(
        df,
        selected_proj,
        selected_proc,
        selected_user)

    if selected_graph:
        graph_content = get_graph_content(df)

    # Get the table data
    records = df.reset_index().to_dict('records')

    # Format records
    for r in records:
         # Make log a link
        _link = r['IDLINK']
        _text = r['ID']
        r['ID'] = f'[{_text}]({_link})'

    # Count how many rows are in the table
    if len(records) > 1:
        rowcount = '{} rows'.format(len(records))
    else:
        rowcount = ''

    return [proc, proj, user, records, graph_content, rowcount, rowcount]



import logging
import os
import pandas as pd

from ....garjus import Garjus


logger = logging.getLogger('dashboard.analyses.data')


def get_filename():
    datadir = f'{Garjus.userdir()}/DATA'
    filename = f'{datadir}/analysesdata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def run_refresh(filename, projects):
    df = get_data(projects)

    save_data(df, filename)

    return df


def load_options():
    garjus = Garjus()
    proj_options = garjus.projects()

    return proj_options


def load_data(projects, refresh=False):
    filename = get_filename()

    if refresh or not os.path.exists(filename):
        run_refresh(filename, projects)

    logger.info('reading data from file:{}'.format(filename))
    return read_data(filename)


def read_data(filename):
    df = pd.read_pickle(filename)
    return df


def save_data(df, filename):
    # save to cache
    df.to_pickle(filename)


def get_data(projects):
    g = Garjus()

    if not g.rcq_enabled():
        logger.debug('rcq not enabled, no analyses data')
        return pd.DataFrame(columns=g.column_names('analyses'))

    # Load
    df = g.analyses(projects, download=False)

    # Pad with zeros
    df['ID'] = df['ID'].astype(str).str.zfill(3)

    df['OUTPUTLINK'] = g.xnat_host() + \
        '/data/projects/' + \
        df['PROJECT'] + \
        '/resources/' + \
        df['OUTPUT'] + \
        '/files'

    df['LOGLINK'] = g.xnat_host() + \
        '/data/projects/' + \
        df['PROJECT'] + \
        '/resources/' + \
        df['OUTPUT'] + \
        '/files/' + \
        df['OUTPUT'] + \
        '.txt'

    df['PDFLINK'] = g.xnat_host() + \
        '/data/projects/' + \
        df['PROJECT'] + \
        '/resources/' + \
        df['OUTPUT'] + \
        '/files/report.pdf'

    df['PBSLINK'] = g.xnat_host() + \
        '/data/projects/' + \
        df['PROJECT'] + \
        '/resources/' + \
        df['OUTPUT'] + \
        '/files/' + \
        df['OUTPUT'] + \
        '.slurm'

    return df


def filter_data(df, time=None):
    # Filter
    if time:
        pass

    return df



"""dashboard home"""
import logging

import pandas as pd
import plotly
import plotly.graph_objs as go
import plotly.subplots
from dash import Input, Output, callback, dcc, html, dash_table as dt
import dash_bootstrap_components as dbc

from .. import utils
from ..shared import STATUS2RGB
from . import data
from .. import queue, activity

logger = logging.getLogger('dashboard.hub')


# Project table with Processing, Automations, Reports

# Bar graph of Queue
# Bar graph of Issues
# Bar graph of Activity

# Analyses graph by status? or list of active?

COMPLETE2EMO = {'0': '🔴', '1': '🟡', '2': '🟢'}


def _reports_graph(df):
    df = df.sort_values('PDF')
    dfp = df.pivot_table(
        index='TYPE',
        values='VIEW',
        columns=['PROJECT'],
        aggfunc='last',
        fill_value='')

    dfp = dfp.map(lambda x: f'[📊]({x})' if x.startswith('http') else x)
    dfp = dfp.reset_index()
    columns = dfp.columns
    records = dfp.to_dict('records')

    # Format columns
    columns = utils.make_columns(columns)
    for i, c in enumerate(columns):
        if c['name'] not in ['TYPE']:
            columns[i]['type'] = 'text'
            columns[i]['presentation'] = 'markdown'

    return [
        dt.DataTable(
            columns=columns,
            data=records,
            filter_action='none',
            page_action='none',
            sort_action='none',
            id='datatable-hub-reports',
            style_cell={
                'textAlign': 'center',
                'width': '10px',
                'height': 'auto',
            },
            style_header={
                'fontWeight': 'bold',
                'padding': '1px 1px 0px 1px',
            },
            css=[
                dict(selector="p", rule="margin: 0; text-align: center;"),
                dict(selector="a", rule="text-decoration: none;"),
            ],
            fill_width=False,
        )]


def _processing_graph(df):

    df['STATUS'] = df['COMPLETE'].map(COMPLETE2EMO).fillna('')

    dfp = df.pivot_table(
        index='TYPE',
        values='STATUS',
        columns=['PROJECT'],
        aggfunc=lambda x: x.mode().iat[0],
        fill_value='')

    dfp = dfp.reset_index()
    columns = dfp.columns
    records = dfp.to_dict('records')

    return [
        dt.DataTable(
            columns=utils.make_columns(columns),
            data=records,
            filter_action='none',
            page_action='none',
            sort_action='none',
            id='datatable-hub-processing',
            style_table={
                'overflowY': 'scroll',
                'overflowX': 'scroll',
            },
            style_cell={
                'textAlign': 'center',
                'width': '10px',
                'height': 'auto',
            },
            style_header={
                'fontWeight': 'bold',
                'padding': '1px 1px 0px 1px',
            },
            fill_width=False,
        )]


def _automations_graph(df):

    df['STATUS'] = df['COMPLETE'].map(COMPLETE2EMO).fillna('')

    dfp = df.pivot_table(
        index='TYPE',
        values='STATUS',
        columns=['PROJECT'],
        aggfunc=lambda x: x.mode().iat[0],
        fill_value='')

    dfp = dfp.reset_index()
    columns = dfp.columns
    records = dfp.to_dict('records')

    return [
        dt.DataTable(
            columns=utils.make_columns(columns),
            data=records,
            filter_action='none',
            page_action='none',
            sort_action='none',
            id='datatable-hub-automations',
            style_cell={
                'textAlign': 'center',
                'width': '10px',
                'height': 'auto',
            },
            style_header={
                'fontWeight': 'bold',
                'padding': '1px 1px 0px 1px',
            },
            fill_width=False,
        )]


def _queue_graph(df):
    if df.empty:
        return [html.P('Nothing in the queue for selected projects.', className='text-center')]

    status2rgb = {k: STATUS2RGB[k] for k in queue.STATUSES}

    # Make a 1x1 figure
    fig = plotly.subplots.make_subplots(rows=1, cols=1)

    dfp = pd.pivot_table(
        df,
        index='PROCTYPE',
        values='LABEL',
        columns=['STATUS'],
        aggfunc='count',
        fill_value=0)

    for status, color in status2rgb.items():
        ydata = sorted(dfp.index)
        if status not in dfp:
            continue
        else:
            xdata = dfp[status]

        fig.append_trace(
            go.Bar(
                x=xdata,
                y=ydata,
                name='{} ({})'.format(status, sum(xdata)),
                marker=dict(color=color),
                opacity=0.9, orientation='h'),
            1,
            1
        )

    fig['layout'].update(barmode='stack', showlegend=True)

    graph = dcc.Graph(figure=fig)

    return [graph]


def _issues_graph(df):
    if df.empty:
        return [html.P('No issues for selected projects.', className='text-center')]

    STATUSES = ['FAIL', 'COMPLETE', 'PASS', 'UNKNOWN']
    status2rgb = {k: STATUS2RGB[k] for k in STATUSES}
    content = []

    # Make a 1x1 figure
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    # Draw bar for each status, these will be displayed in order
    dfp = pd.pivot_table(
        df,
        index='CATEGORY',
        values='LABEL',
        columns=['STATUS'],
        aggfunc='count',
        fill_value=0)

    for status, color in status2rgb.items():
        ydata = sorted(dfp.index)
        if status not in dfp:
            xdata = [0] * len(dfp.index)
        else:
            xdata = dfp[status]

        fig.append_trace(go.Bar(
            x=ydata,
            y=xdata,
            name='{} ({})'.format(status, sum(xdata)),
            marker=dict(color=color),
            opacity=0.9), 1, 1)

    # Customize figure
    fig['layout'].update(barmode='stack', showlegend=False)

    content.append(dcc.Graph(figure=fig))

    return content


def _activity_graph(df):
    if df.empty:
        return [html.P('No recent activity for selected projects', className='text-center')]

    status2rgb = {k: STATUS2RGB[k] for k in activity.STATUSES}
    content = []

    # Make a 1x1 figure
    fig = plotly.subplots.make_subplots(rows=1, cols=1)
    fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))

    # Draw bar for each status, these will be displayed in order
    dfp = pd.pivot_table(
        df, index='CATEGORY', values='ID', columns=['STATUS'],
        aggfunc='count', fill_value=0)

    for status, color in status2rgb.items():
        ydata = sorted(dfp.index)
        if status not in dfp:
            xdata = [0] * len(dfp.index)
        else:
            xdata = dfp[status]

        fig.append_trace(go.Bar(
            x=ydata,
            y=xdata,
            name='{} ({})'.format(status, sum(xdata)),
            marker=dict(color=color),
            opacity=0.9), 1, 1)

    # Customize figure
    fig['layout'].update(barmode='stack', showlegend=False)

    graph = dcc.Graph(figure=fig)

    content.append(graph)

    return content


def get_content():
    '''Get page content.'''

    # We use the dbc grid layout with rows and columns, rows are 12 units wide
    content = [
        dbc.Row([
            dbc.Col(dbc.Button('Refresh', id='button-hub-refresh')),
        ]),
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-hub-proj',
                    multi=True,
                    placeholder='Select Project(s)',
                ),
                width=3,
            ),
        ),
        dbc.Spinner([
            dbc.Row([
                dbc.Col(
                    html.H5('Activity', className='text-center'), width=3),
                dbc.Col(html.H5('Queue', className='text-center'), width=6),
                dbc.Col(html.H5('Issues'), className='text-center', width=3),
            ]),
            dbc.Row([
                dbc.Col(
                    html.Div(id='div-hub-activity', children=[]), width=3,
                ),
                dbc.Col(
                    html.Div(id='div-hub-queue', children=[]), width=6,
                ),
                dbc.Col(
                    html.Div(id='div-hub-issues', children=[]), width=3,
                ),
            ]),
        ]),
        dbc.Row([dbc.Col(html.H5('Reports'))]),
        dbc.Row([
            dbc.Col(
                html.Div(
                    id='div-hub-reports',
                    children=[],
                    style={'margin-bottom': '2em'}))]),
        dbc.Row([dbc.Col(html.H5('Processing'))]),
        dbc.Row([
            dbc.Col(
                html.Div(
                    id='div-hub-processing',
                    children=[],
                    style={'margin-bottom': '2em'},
                ),
                width=12
            )
        ]),
        #dbc.Row([dbc.Col(html.Label('Automations:TBD'))]),
    ]

    return content


@callback(
    [
     Output('dropdown-hub-proj', 'options'),
     Output('div-hub-processing', 'children'),
     Output('div-hub-reports', 'children'),
     Output('div-hub-queue', 'children'),
     Output('div-hub-issues', 'children'),
     Output('div-hub-activity', 'children'),
     ],
    [
     Input('button-hub-refresh', 'n_clicks'),
     Input('dropdown-hub-proj', 'value'),
    ],
)
def update_hub(n_clicks, selected_proj):
    refresh = False

    logger.debug('update_hub')

    if utils.was_triggered('button-hub-refresh'):
        # Refresh data if refresh button clicked
        logger.debug('refresh-hub:clicks={}'.format(n_clicks))
        refresh = True

    # Load datas
    queue_data = data._get_queue_data(refresh=refresh)
    issues_data = data._get_issues_data(refresh=refresh)
    act_data = data._get_activity_data(refresh=refresh)
    proc_data = data._get_processing_data()
    reports_data = data._get_reports_data(refresh=refresh)

    # Get options for dropdowns berfore filtering
    proj_options = data._load_options(proc_data)
    proj = utils.make_options(proj_options)

    # Filter data
    if selected_proj:
        proc_data = proc_data[proc_data.PROJECT.isin(selected_proj)]
        queue_data = queue_data[queue_data.PROJECT.isin(selected_proj)]
        act_data = act_data[act_data.PROJECT.isin(selected_proj)]
        issues_data = issues_data[issues_data.PROJECT.isin(selected_proj)]
        reports_data = reports_data[reports_data.PROJECT.isin(selected_proj)]

    # Make graphs/tables
    queue_graph = _queue_graph(queue_data)
    issues_graph = _issues_graph(issues_data)
    act_graph = _activity_graph(act_data)
    proc_graph = _processing_graph(proc_data)
    reports_graph = _reports_graph(reports_data)

    # Return table, figure, dropdown options
    logger.debug('update_hub:returning data')

    return [proj, proc_graph, reports_graph, queue_graph, issues_graph, act_graph]



"""dash index page."""
import logging

from dash import html
import dash_bootstrap_components as dbc

from .app import app
from .pages import hub
from .pages import qa
from .pages import activity
from .pages import issues
from .pages import queue
from .pages import stats
from .pages import analyses
from .pages import processors
from .pages import reports
from ..garjus import Garjus


logger = logging.getLogger('garjus.dashboard.content')


def _redcap_found():
    return Garjus.redcap_found()


def _xnat_found():
    return Garjus.xnat_found()


def _rcq_found():
    return Garjus.rcq_found()


def _footer_content(include_logout=False):
    content = []

    content.append(html.Hr())

    if include_logout:
        content.append(
            html.Div([
                dbc.Row([
                    dbc.Col(
                        html.A(
                            "garjus",
                            href='https://github.com/ccmvumc/garjus',
                            target="_blank",
                        ),
                    ),
                    dbc.Col(
                        html.A('xnat', href='https://xnat.vanderbilt.edu/xnat'),
                    ),
                    dbc.Col(
                        html.A('logout', href='../logout'),
                    ),
                ]),
                ],
                style={'textAlign': 'center'},
            )
        )
    else:
        content.append(
            html.Div([
                html.A(
                    "garjus",
                    href='https://github.com/ccmvumc/garjus',
                    target="_blank",
                )
                ],
                style={'textAlign': 'center'}
            )
        )

    return content


def get_content(include_logout=False, demo=False):

    if demo:
        return demo_content()

    #has_xnat = _xnat_found()
    has_xnat = True
    has_redcap = _redcap_found()
    has_rcq = _rcq_found()
    tabs = ''
    content = ''

    logger.debug(f'{has_xnat=}, {has_redcap=}')

    if has_xnat and has_redcap and has_rcq:
        # include all tabs
        tabs = dbc.Tabs([
            #dbc.Tab(
            #    label='Home',
            #    tab_id='tab-home',
            #    children=hub.get_content(),
            #),
            dbc.Tab(
                label='QA',
                tab_id='tab-qa',
                children=qa.get_content(),
            ),
            dbc.Tab(
                label='Issues',
                tab_id='tab-issues',
                children=issues.get_content(),
             ),
            dbc.Tab(
                label='Queue',
                tab_id='tab-queue',
                children=queue.get_content(),
            ),
            dbc.Tab(
                label='Activity',
                tab_id='tab-activity',
                children=activity.get_content(),
            ),
            dbc.Tab(
                label='Stats',
                tab_id='tab-stats',
                children=stats.get_content(),
            ),
            dbc.Tab(
                label='Processors',
                tab_id='tab-processors',
                children=processors.get_content(),
            ),
            dbc.Tab(
                label='Reports',
                tab_id='tab-reports',
                children=reports.get_content(),
            ),
            dbc.Tab(
                label='Analyses',
                tab_id='tab-analyses',
               children=analyses.get_content(),
            )
            ],
            active_tab="tab-qa",
        )
    elif has_xnat and has_redcap and not has_rcq:
        # include all tabs
        tabs = dbc.Tabs([
            dbc.Tab(
                label='QA',
                tab_id='tab-qa',
                children=qa.get_content(),
            ),
            dbc.Tab(
                label='Issues',
                tab_id='tab-issues',
                children=issues.get_content(),
            ),
            dbc.Tab(
                label='Activity',
                tab_id='tab-activity',
                children=activity.get_content(),
            ),
            dbc.Tab(
                label='Stats',
                tab_id='tab-stats',
                children=stats.get_content(),
            ),
            dbc.Tab(
                label='Reports',
                tab_id='tab-reports',
                children=reports.get_content(),
            )],
            active_tab="tab-qa",
        )
    elif has_xnat and not has_redcap and has_rcq:
        # include all tabs
        tabs = dbc.Tabs([
            dbc.Tab(
                label='QA',
                tab_id='tab-qa',
                children=qa.get_content(),
            ),
            dbc.Tab(
                label='Processors',
                tab_id='tab-processors',
                children=processors.get_content(),
            ),
            dbc.Tab(
                label='Analyses',
                tab_id='tab-analyses',
               children=analyses.get_content(),
            ),
            dbc.Tab(
                label='Queue',
                tab_id='tab-queue',
                children=queue.get_content(),
            )],
            active_tab="tab-qa",
        )
    elif has_xnat and not has_redcap:
        tabs = html.Div(qa.get_content())
    elif has_redcap and not has_xnat:
        tabs = dbc.Tabs([
            dbc.Tab(
                label='Issues',
                tab_id='tab-issues',
                children=issues.get_content(),
            ),
            dbc.Tab(
                label='Queue',
                tab_id='tab-queue',
                children=queue.get_content(),
            ),
            dbc.Tab(
                label='Activity',
                tab_id='tab-activity',
                children=activity.get_content(),
            ),
            dbc.Tab(
                label='Processors',
                tab_id='tab-processors',
                children=processors.get_content(),
            ),
            dbc.Tab(
                label='Reports',
                tab_id='tab-reports',
                children=reports.get_content(),
            ),
            dbc.Tab(
                label='Analyses',
                tab_id='tab-analyses',
                children=analyses.get_content(),
            ),
        ])

    footer_content = _footer_content(include_logout)

    content = html.Div(
        className='dbc',
        style={'marginLeft': '20px', 'marginRight': '20px'},
        children=[
            html.Div(id='report-content', children=[tabs]),
            html.Div(id='footer-content', children=footer_content)
    ])

    return content


def demo_content():
    tabs = dbc.Tabs([
        dbc.Tab(
            label='QA',
            tab_id='tab-qa',
            children=qa.get_content(),
        ),
        #dbc.Tab(
        #    label='Issues',
        #    tab_id='tab-issues',
        #    children=issues.get_content(),
        # ),
        #dbc.Tab(
        #    label='Queue',
        #    tab_id='tab-queue',
        #    children=queue.get_content(),
        # ),
        #dbc.Tab(
        #    label='Activity',
        #    tab_id='tab-activity',
        #    children=activity.get_content(),
        #),
        #dbc.Tab(
        #    label='Stats',
        #    tab_id='tab-stats',
        #    children=stats.get_content(),
        #),
        #dbc.Tab(
        #    label='Processors',
        #    tab_id='tab-processors',
        #    children=processors.get_content(),
        #),
        #dbc.Tab(
        #    label='Reports',
        #    tab_id='tab-reports',
        #    children=reports.get_content(),
        #),
        #dbc.Tab(
        #    label='Analyses',
        #    tab_id='tab-analyses',
        #   children=analyses.get_content(),
        #)
    ], active_tab="tab-qa")

    footer_content = _footer_content()

    content = html.Div(
        className='dbc',
        style={'marginLeft': '20px', 'marginRight': '20px'},
        children=[
            html.Div(id='report-content', children=[tabs]),
            html.Div(id='footer-content', children=footer_content)
    ])

    return content



import redcap
import os
import logging


def download_named_file(
    project,
    record_id,
    field_id,
    outdir,
    event_id=None,
    repeat_id=None
):
    # Get the file contents from REDCap
    try:
        (cont, hdr) = project.export_file(
            record=record_id,
            field=field_id,
            event=event_id,
            repeat_instance=repeat_id)

        if cont == '':
            raise redcap.RedcapError
    except redcap.RedcapError as err:
        logging.error(f'downloading file:{err}')
        return None

    # Save contents to local file
    filename = os.path.join(outdir, hdr['name'])
    try:
        with open(filename, 'wb') as f:
            f.write(cont)

        return filename
    except FileNotFoundError as err:
        logging.error(f'file not found:{filename}:{err}')
        return None


def download_file(
    project,
    record_id,
    field_id,
    filename,
    event_id=None,
    repeat_id=None
):
    # Get the file contents from REDCap
    try:
        (cont, hdr) = project.export_file(
            record=record_id,
            field=field_id,
            event=event_id,
            repeat_instance=repeat_id)

        if cont == '':
            raise redcap.RedcapError
    except redcap.RedcapError as err:
        logging.error(f'downloading file:{err}')
        return None

    # Save contents to local file
    try:
        with open(filename, 'wb') as f:
            f.write(cont)

        return filename
    except FileNotFoundError as err:
        logging.error(f'file not found:{filename}:{err}')
        return None


def upload_file(
    project,
    record_id,
    field_id,
    filename,
    event_id=None,
    repeat_id=None
):
    with open(filename, 'rb') as f:
        return project.import_file(
            record=record_id,
            field=field_id,
            file_name=os.path.basename(filename),
            event=event_id,
            repeat_instance=repeat_id,
            file_object=f)


def get_redcap(project_id=None, key_file=None, api_url=None, api_key=None):
    # Check for overrides in environment vars
    api_url = os.environ.get('REDCAP_API_URL', api_url)
    key_file = os.environ.get('REDCAP_API_KEYFILE', key_file)

    if not api_url:
        api_url = 'https://redcap.vumc.org/api/'

    if not api_key:
        # key not specified so we read it from file

        if not key_file:
            # no key file specified so we use the default location
            key_file = os.path.join(os.path.expanduser('~'), '.redcap.txt')

        # Load from the key file
        if project_id:
            api_key = get_projectkey(project_id, key_file)

    if not api_key:
        raise Exception('api key not found in file or arguments')

    return redcap.Project(api_url, api_key)


def get_projectkey(project_id, key_file):
    # Load the dictionary
    d = {}
    with open(key_file) as f:
        for line in f:
            if line == '':
                continue

            try:
                (i, k, n) = line.strip().split(',')
                d[i] = k
            except Exception:
                pass

    # Return the key id for given project id
    return d.get(project_id, None)


def get_projectid(projectname, keyfile):
    # Load the dictionary mapping name to id
    d = {}
    with open(keyfile) as f:
        for line in f:
            if line == '':
                continue
            try:
                (i, k, n) = line.strip().split(',')
                # Map name to id
                d[n] = i
            except Exception:
                pass
    # Return the project id for given project name
    return d.get(projectname, None)


def get_main_redcap():
    api_url = 'https://redcap.vumc.org/api/'
    keyfile = os.path.join(os.path.expanduser('~'), '.redcap.txt')

    # Check for overrides in environment vars
    api_url = os.environ.get('REDCAP_API_URL', api_url)
    keyfile = os.environ.get('REDCAP_API_KEYFILE', keyfile)

    project_id = get_projectid('main', keyfile)
    api_key = get_projectkey(project_id, keyfile)

    if not api_key:
        return None

    return redcap.Project(api_url, api_key)


def get_rcq_redcap():
    api_url = 'https://redcap.vumc.org/api/'
    keyfile = os.path.join(os.path.expanduser('~'), '.redcap.txt')

    # Check for overrides in environment vars
    api_url = os.environ.get('REDCAP_API_URL', api_url)
    keyfile = os.environ.get('REDCAP_API_KEYFILE', keyfile)

    project_id = get_projectid('rcq', keyfile)
    api_key = get_projectkey(project_id, keyfile)

    if not api_key:
        return None

    return redcap.Project(api_url, api_key)


def get_identifier_redcap():
    api_url = 'https://redcap.vumc.org/api/'
    keyfile = os.path.join(os.path.expanduser('~'), '.redcap.txt')

    # Check for overrides in environment vars
    api_url = os.environ.get('REDCAP_API_URL', api_url)
    keyfile = os.environ.get('REDCAP_API_KEYFILE', keyfile)

    project_id = get_projectid('identifier', keyfile)
    api_key = get_projectkey(project_id, keyfile)

    return redcap.Project(api_url, api_key)


def match_repeat(rc, record_id, repeat_name, match_field, match_value):
    # Load potential matches
    records = rc.export_records(records=[record_id])

    # Find records with matching vaue
    matches = [x for x in records if x[match_field] == match_value]

    # Return ids of matches
    return [x['redcap_repeat_instance'] for x in matches]


def field2events(project, field_id):
    events = []

    try:
        _form = [x['form_name'] for x in project.metadata if x['field_name'] == field_id][0]
        events = [x['unique_event_name'] for x in project.export_instrument_event_mappings() if x['form'] == _form]
    except IndexError:
        events = []

    return events


def secondary_map(project):
    def_field = project.def_field
    sec_field = secondary(project)

    if not sec_field:
        return {}

    # Get the secondary values
    rec = project.export_records(fields=[def_field, sec_field])

    # Build the map
    id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}

    return id2subj

def secondary(project):
    return project.export_project_info()['secondary_unique_field']



import logging

import pandas as pd
from dash import dcc, html, dash_table as dt, Input, Output, callback
import dash_bootstrap_components as dbc

from .. import utils
from . import data

logger = logging.getLogger('dashboard.analyses')


#COMPLETE2EMO = {'0': '🔴', '1': '🟡', '2': '🟢'}

# command line examples for interacting with analyses
TIPS_MARKDOWN = '''
    &nbsp;

    ### Analyses Tips:

    &nbsp;

    To download all input files for analysis number NUM to folder INPUTS for project NAME, at command-line enter:  
    `garjus getinputs -p NAME NUM INPUTS`

    For example, to download inputs for analysis 1 from ProjectA to a local folder named INPUTS, enter:  
    `garjus getinputs -p ProjectA 1 ./INPUTS`

    &nbsp;

    To download the output zip for analysis number NUM to folder OUTPUTS for project NAME, at command-line enter:  
    `garjus getoutputs -p NAME NUM OUTPUTS`

    For example, to download outputs for analysis 1 from ProjectA to a local folder named OUTPUTS, enter:  
    `garjus getoutputs -p ProjectA 1 ./OUTPUTS`

    &nbsp;

    To run an analysis by downloading the inputs locally and saving the outputs locally:  
    `garjus run -p NAME NUM OUTPUTS.zip`

    For example, to run analysis 3 for project ProjectA and save outputs to zip:  
    `garjus run -p ProjectA 3 ProjectA_3_OUTPUTS.zip`

    &nbsp;

    To update analyses for a project, enter:  
    `garjus update analyses -p ProjectA`

    This will update each analyses by creating the inputs if already done,
    and if outputs does not exist, then it will run the analysis to create the outputs,
    upload a zip to project resources on XNAT, and finally create a link in the analysis record.  

    &nbsp;

'''

COLUMNS = [
    'PROJECT',
    'ID',
    'NAME',
    'EDIT',
    'STATUS',
    'PBS',
    'PDF',
    'LOG',
    'OUTPUT',
    'SUBJECTS',
    'INVESTIGATOR',
    'PROCESSOR',
    'NOTES'
]


def get_content():
    columns = utils.make_columns(COLUMNS)

    # Format columns with links as markdown text
    for i, c in enumerate(columns):
        if c['name'] in ['OUTPUT', 'EDIT', 'INPUT', 'DATA', 'PROCESSOR', 'LOG', 'PDF', 'PBS']:
            columns[i]['type'] = 'text'
            columns[i]['presentation'] = 'markdown'

    content = [
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-analyses-proj',
                    multi=True,
                    placeholder='Select Project(s)',
                ),
                width=3,
            ),
        ),
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-analyses-lead',
                    multi=True,
                    placeholder='Select Lead Investigator(s)',
                ),
                width=3,
            ),
        ),
        dbc.Row(
            dbc.Col(
                dcc.Dropdown(
                    id='dropdown-analyses-status',
                    multi=True,
                    placeholder='Select Status',
                ),
                width=3,
            ),
        ),
        dbc.Spinner(id="loading-analyses-table", children=[
            dbc.Label('Loading...', id='label-analyses-rowcount1'),
        ]),
        dt.DataTable(
            columns=columns,
            data=[],
            page_action='none',
            sort_action='none',
            id='datatable-analyses',
            style_cell={
                'textAlign': 'center',
                'height': 'auto',
                'padding': '15px 5px 15px 5px',
            },
            style_header={
                'fontWeight': 'bold',
            },
            style_cell_conditional=[
                {'if': {'column_id': 'NAME'}, 'textAlign': 'left'},
            ],
            # Aligns the markdown cells, both vertical and horizontal, and 
            # prevent extra underlines around links
            css=[
                dict(selector="p", rule="margin: 0; text-align: center"),
                dict(selector="a", rule="text-decoration: none;"),
            ],
        ),
        html.Label('0', id='label-analyses-rowcount2'),
        #dcc.Markdown(TIPS_MARKDOWN)
    ]

    return content


def load_analyses(projects=[]):

    if projects is None:
        projects = []

    return data.load_data(projects, refresh=True)


@callback(
    [
     Output('dropdown-analyses-proj', 'options'),
     Output('dropdown-analyses-lead', 'options'),
     Output('dropdown-analyses-status', 'options'),
     Output('datatable-analyses', 'data'),
     Output('label-analyses-rowcount1', 'children'),
     Output('label-analyses-rowcount2', 'children'),
    ],
    [
     Input('dropdown-analyses-proj', 'value'),
     Input('dropdown-analyses-lead', 'value'),
     Input('dropdown-analyses-status', 'value'),
    ])
def update_analyses(
    selected_proj,
    selected_lead,
    selected_status,
):
    logger.debug('update_all')

    # Load selected data with refresh if requested
    df = load_analyses(selected_proj)

    # Truncate NOTES
    if 'NOTES' in df:
        df['NOTES'] = df['NOTES'].str.slice(0, 20)

    # Count SUBJECTS list
    df.loc[df['SUBJECTS'].str.len() > 0, 'SUBJECTS'] = 'n=' + df['SUBJECTS'].str.split(r'[,\n\s]+', regex=True, expand=False).agg(len).astype(str)

    # Change blanks to asterisk
    df.loc[df['SUBJECTS'].str.len() == 0, 'SUBJECTS'] = '*'

    # Get options
    proj_options = data.load_options()
    lead_options = sorted(df['INVESTIGATOR'].unique())
    status_options = sorted(df['STATUS'].unique())
    logger.debug(f'loaded options:{proj_options}:{lead_options}')

    proj = utils.make_options(proj_options)
    lead = utils.make_options(lead_options)
    status = utils.make_options(status_options)

    # Filter data based on dropdown values
    df = data.filter_data(df)

    if selected_lead:
        df = df[df['INVESTIGATOR'].isin(selected_lead)]

    if selected_status:
        df = df[df['STATUS'].isin(selected_status)]

    #df['COMPLETE'] = df['COMPLETE'].map(COMPLETE2EMO).fillna('?')

    # Get the table data as one row per assessor
    records = df.reset_index().to_dict('records')

    # Format records
    for r in records:
        # Make edit a link
        _link = r['EDIT']
        _text = 'edit'
        r['EDIT'] = f'[{_text}]({_link})'

         # Make log a link
        _link = r['LOGLINK']
        r['LOG'] = f'[📄]({_link})'

         # Make pdf a link
        _link = r['PDFLINK']
        r['PDF'] = f'[📊]({_link})'

        # Make pbs a link
        _link = r['PBSLINK']
        r['PBS'] = f'[📋]({_link})'

        # Make a link
        if not r['OUTPUT']:
            pass
        elif r['OUTPUTLINK']:
            _link = r['OUTPUTLINK']
            _text = r['OUTPUT']
            r['OUTPUT'] = f'[{_text}]({_link})'
        elif '/' in r['OUTPUT']:
            _link = r['OUTPUT']
            _text = r['OUTPUT'].rsplit('/', 2)[1]
            r['OUTPUT'] = f'[{_text}]({_link})'

        # Make a link
        if not r['PROCESSOR']:
            pass
        elif '/' in r['PROCESSOR']:
            try:
                p = r['PROCESSOR'].replace(':', '/').split('/')
                if len(p) == 4:
                    _link = f'https://github.com/{p[0]}/{p[1]}/tree/{p[2]}/processors/{p[3]}'
                elif len(p) == 3:
                    _link = f'https://github.com/{p[0]}/{p[1]}/tree/{p[2]}'
                else:
                    _link = f'https://github.com/{p}'

                _text = r['PROCESSOR']
                r['PROCESSOR'] = f'[{_text}]({_link})'
            except Exception as err:
                logger.error(f'failed to parse processor:{r["PROCESSOR"]}')

    # Count how many rows are in the table
    rowcount = '{} rows'.format(len(records))

    return [proj, lead, status, records, rowcount, rowcount]



import logging

import pandas as pd
import plotly
import plotly.graph_objs as go
import plotly.subplots
from dash import Input, Output, callback,dcc, html, dash_table as dt
import dash_bootstrap_components as dbc

from .. import utils
from . import data
from ..shared import GWIDTH


logger = logging.getLogger('dashboard.stats')


HIDECOLS = [
    'ASSR',
    'PROJECT',
    'SUBJECT',
    'SESSION',
    'SESSIONLINK',
    'SESSTYPE',
    'SITE',
    'DATE',
    'PROCTYPE'
]


def _plottable(var):
    if pd.api.types.is_numeric_dtype(var):
        return True

    try:
        _ = var.astype(float)
        return True
    except Exception:
        pass

    try:
        _ = var.str.strip('%').astype(float)
        return True
    except Exception:
        return False


def get_graph_content(df, selected_pivot):
    tabs_content = []
    tab_value = 0
    hidecols = HIDECOLS
    logger.debug('get_stats_figure')

    # Check for empty data
    if len(df) == 0:
        logger.debug('empty data, using empty figure')
        _txt = 'Choose Project(s) then Type(s) to load stats'
        return [dcc.Tab(label='', value='0', children=[html.Div(
            html.P(_txt, style={'text-align': 'center'}),
            style={'padding':'100px', 'height': '100px', 'width': '700px'})])]

    # Filter var list to only include those that have data
    var_list = [x for x in df.columns if not pd.isnull(df[x]).all()]

    # Hide some columns
    var_list = [x for x in var_list if x not in hidecols]

    # Hide more columns
    var_list = [x for x in var_list if not (
        x.endswith('_pctused') or 
        x.endswith('_voltot')) or
        x.endswith('_volused')
    ]

    # Filter var list to only stats can be plotted as float
    var_list = [x for x in var_list if _plottable(df[x])]

    # Append the tab
    _label = 'ALL'
    _graph = get_stats_graph(df, var_list)
    _tab = dcc.Tab(label=_label, value=str(tab_value), children=[_graph])
    tabs_content.append(_tab)

    # other pivots
    pivots = ['SITE', 'PROJECT', 'SESSTYPE']
    for p in pivots:

        if len(df[p].unique()) <= 1:
            continue

        tab_value += 1
        _label = 'By ' + p
        _graph = get_stats_graph(df, var_list, p)
        _tab = dcc.Tab(label=_label, value=str(tab_value), children=[_graph])
        tabs_content.append(_tab)

    # Return the tabs
    return tabs_content


def get_stats_graph(df, var_list, pivot=None):
    box_width = 250
    min_box_count = 4

    logger.debug(f'get_graph_tab:{pivot}')

    # Determine how many boxplots we're making, depends on how many vars, use
    # minimum so graph doesn't get too small
    box_count = len(var_list)
    if box_count < min_box_count:
        box_count = min_box_count

    graph_width = box_width * box_count

    # Horizontal spacing cannot be greater than (1 / (cols - 1))
    hspacing = 1 / (box_count * 4)

    # Make the figure with 1 row and a column for each var we are plotting
    var_titles = [x[:22] for x in var_list]
    fig = plotly.subplots.make_subplots(
        rows=1,
        cols=box_count,
        horizontal_spacing=hspacing,
        subplot_titles=var_titles)

    # Draw boxplots by adding traces to figure
    for i, var in enumerate(var_list):
        _row = 1
        _col = i + 1

        # Create boxplot for this var and add to figure
        logger.debug(f'plotting var:{var}')

        # Create boxplot for this var and add to figure
        if pivot:
            _xvalues = df[pivot]
        else:
            _xvalues = None

        fig.append_trace(
            go.Box(
                y=df[var].astype(str).str.strip('%').astype(float),
                x=_xvalues,
                boxpoints='all',
                text=df['ASSR'],
                boxmean=True,
            ),
            _row,
            _col)

        if var.startswith('con_') or var.startswith('inc_'):
            fig.update_yaxes(range=[-1, 1], autorange=False)
        else:
            fig.update_yaxes(autorange=True)
            pass

    # Customize figure to hide legend and fit the graph
    fig.update_layout(
        showlegend=False,
        autosize=False,
        width=graph_width,
        margin=dict(l=20, r=40, t=40, b=80, pad=0))

    if not pivot:
        fig.update_xaxes(showticklabels=False)  # hide all the xticks

    # Build the tab
    # We set the graph to overflow and then limit the size to 1000px, this
    # makes the graph stay in a scrollable section
    graph = html.Div(
        dcc.Graph(figure=fig, style={'overflow': 'scroll'}),
        style={'width': f'{GWIDTH}px'})

    # Return the graph
    return graph


def get_content():
    proj, _ = data.load_options()
    if not proj:
        logger.debug('no projects for stats')
        return None

    content = [
        dbc.Button(
            'Refresh Data',
            id='button-stats-refresh',
            outline=True,
            color='primary',
            size='sm',
        ),
        dbc.Spinner(id="loading-stats", children=[
            html.Div(dcc.Tabs(
                id='tabs-stats',
                value='0',
                children=[],
                vertical=True))]
        ),
        dbc.Row(
            dbc.Col(
                dbc.Stack([
                    dcc.Dropdown(
                        id='dropdown-stats-time',
                        options=[
                            {'label': 'all time', 'value': 'ALL'},
                            {'label': '1 day', 'value': '1day'},
                            {'label': '1 week', 'value': '7day'},
                            {'label': '1 month', 'value': '30day'},
                            {'label': '1 year', 'value': '365day'}],
                        value='ALL'
                    ),
                    dcc.Dropdown(
                        id='dropdown-stats-proj',
                        multi=True,
                        placeholder='Select Project(s)'
                    ),
                    dcc.Dropdown(
                        id='dropdown-stats-proc',
                        multi=True,
                        placeholder='Select Type(s)'
                    ),
                    dcc.Dropdown(
                        id='dropdown-stats-sess',
                        multi=True,
                        placeholder='Select Session Type(s)'
                    ),
                ]),
                width=5,
            ),
        ),
        dbc.Row(
            dbc.Col(
                dbc.RadioItems(
                    # Use specific css to make radios look like buttons
                    className="btn-group",
                    inputClassName="btn-check",
                    labelClassName="btn btn-outline-primary",
                    labelCheckedClassName="active",
                    options=[
                        {'label': 'Assessors', 'value': 'assr'},
                        {'label': 'Sessions', 'value': 'sess'},
                        {'label': 'Subjects', 'value': 'subj'}],
                    value='assr',
                    id='radio-stats-pivot',
                    labelStyle={'display': 'inline-block'},
                ),
                align='end',
                width=5,
            ),
        ),
        dbc.Spinner(id="loading-stats-table", children=[
            dbc.Label('Get ready...', id='label-stats-rowcount1'),
        ]),
        dt.DataTable(
            columns=[],
            data=[],
            page_action='none',
            sort_action='native',
            id='datatable-stats',
            style_table={
                'overflowY': 'scroll',
                'overflowX': 'scroll',
            },
            style_cell={
                'textAlign': 'left',
                'padding': '5px 5px 0px 5px',
                'width': '30px',
                'overflow': 'hidden',
                'textOverflow': 'ellipsis',
                'height': 'auto',
                'minWidth': '40',
                'maxWidth': '60'},
            style_header={
                'textAlign': 'center',
                'fontWeight': 'bold',
                'padding': '5px 15px 0px 10px'},
            style_cell_conditional=[
                {'if': {'column_type': 'numeric'}, 'textAlign': 'right'},
                {'if': {'column_id': 'PROJECT'}, 'textAlign': 'center'},
                {'if': {'column_id': 'SUBJECT'}, 'textAlign': 'center'},
                {'if': {'column_id': 'PROCTYPE'}, 'textAlign': 'center'},
                {'if': {'column_id': 'SESSTYPE'}, 'textAlign': 'center'},
                {'if': {'column_id': 'SITE'}, 'textAlign': 'center'},
            ],
            css=[dict(selector="p", rule="margin: 0; text-align: center")],
            export_format='xlsx',
            export_headers='names',
            export_columns='visible',
        ),
        dbc.Label('Get ready...', id='label-stats-rowcount2'),
    ]

    return content


def load_stats(projects=[], refresh=False, filename=None):

    if projects is None:
        projects = []

    return data.load_data(projects, refresh=refresh, filename=filename)


def _subject_pivot(df):
    # Pivot to one row per subject

    index_cols = ['PROJECT', 'SUBJECT', 'SITE']

    level_cols = []

    if len(df.SESSTYPE.unique()) > 1:
        level_cols.append('SESSTYPE')

    if len(df.PROCTYPE.unique()) > 1:
        level_cols.append('PROCTYPE')

    stat_cols = []

    # Drop any duplicates found
    df = df.drop_duplicates()

    # And duplicate proctype for session
    df = df.drop_duplicates(
        subset=['SUBJECT', 'SESSTYPE', 'PROCTYPE'],
        keep='last')

    df = df.drop(columns=['ASSR', 'SESSION', 'SESSIONLINK'])

    # Build list of stat columns
    stat_cols = [x for x in df.columns if (x not in index_cols and x not in ['SESSTYPE', 'PROCTYPE'])]

    # Make the pivot table based on _index, _cols, _vars
    dfp = df.pivot(index=index_cols, columns=level_cols, values=stat_cols)

    if len(level_cols) == 1:
        # Concatenate column levels to get one level with delimiter
        dfp.columns = [f'{c[1]}_{c[0]}' for c in dfp.columns.values]
    elif len(level_cols) == 2:
        # Concatenate column levels to get one level with delimiter
        dfp.columns = [f'{c[2]}_{c[1]}_{c[0]}' for c in dfp.columns.values]

    # Clear the index so all columns are named
    dfp = dfp.dropna(axis=1, how='all')
    dfp = dfp.reset_index()

    return dfp


def _session_pivot(df):
    # Pivot to one row per session
    level_cols = ['PROCTYPE']
    stat_cols = []
    index_cols = ['PROJECT', 'SUBJECT', 'SITE', 'SESSION', 'SESSIONLINK', 'SESSTYPE', 'DATE']

    # Drop any duplicates found
    df = df.drop_duplicates()

    # And drop any duplicate proctypes per session
    df = df.drop_duplicates(
        subset=['SUBJECT', 'SESSTYPE', 'PROCTYPE', 'DATE'],
        keep='last')

    df = df.drop(columns=['ASSR'])

    stat_cols = [x for x in df.columns if (x not in index_cols and x not in level_cols)]

    # Make the pivot table based on _index, _cols, _vars
    dfp = df.pivot(index=index_cols, columns=level_cols, values=stat_cols)

    dfp.columns = [c[0] for c in dfp.columns.values]

    # Clear the index so all columns are named
    dfp = dfp.dropna(axis=1, how='all')
    dfp = dfp.reset_index()

    return dfp


@callback(
    [Output('dropdown-stats-proc', 'options'),
     Output('dropdown-stats-proj', 'options'),
     Output('dropdown-stats-sess', 'options'),
     Output('datatable-stats', 'data'),
     Output('datatable-stats', 'columns'),
     Output('tabs-stats', 'children'),
     Output('label-stats-rowcount1', 'children'),
     Output('label-stats-rowcount2', 'children'),
    ],
    [
     Input('dropdown-stats-proc', 'value'),
     Input('dropdown-stats-proj', 'value'),
     Input('dropdown-stats-sess', 'value'),
     Input('dropdown-stats-time', 'value'),
     Input('radio-stats-pivot', 'value'),
     Input('button-stats-refresh', 'n_clicks'),
    ])
def update_stats(
    selected_proc,
    selected_proj,
    selected_sess,
    selected_time,
    selected_pivot,
    n_clicks
):
    refresh = False

    logger.debug('update_all')

    if utils.was_triggered('button-stats-refresh'):
        # Refresh data if refresh button clicked
        logger.debug('refresh:clicks={}'.format(n_clicks))
        refresh = True

    # Load selected data with refresh if requested
    df = load_stats(selected_proj, refresh=refresh)

    if selected_proj and (df.empty or (sorted(selected_proj) != sorted(df.PROJECT.unique()))):
        # A new project was selected so we force refresh
        logger.debug('new project selected, refreshing')
        df = load_stats(selected_proj, refresh=True)

    # Get options based on selected projects, only show proc for those projects
    proj_options, proc_options = data.load_options(selected_proj)

    logger.debug(f'loaded options:{proj_options}:{proc_options}')

    proj = utils.make_options(proj_options)
    proc = utils.make_options(proc_options)

    # Get session types from unfiltered data
    if not df.empty:
        sess = utils.make_options(df.SESSTYPE.unique())
    else:
        sess = []

    # Filter data based on dropdown values
    df = data.filter_data(df, selected_proc, selected_time, selected_sess)

    # Get the graph content in tabs (currently only one tab)
    tabs = get_graph_content(df, selected_pivot)

    # format floats so they sort in the table
    for c in list(df.columns):
        if _plottable(df[c]):
            df[c] = df[c].astype(str).str.strip('%').astype(float)

    # Apply pivot
    if selected_pivot == 'subj':
        df = _subject_pivot(df)
    elif selected_pivot == 'sess':
        df = _session_pivot(df)

    # Get the records and columns for DataTable
    _cols = [x for x in list(df.columns) if x not in ['SESSIONLINK']]
    columns = utils.make_columns(_cols)
    records = df.reset_index().to_dict('records')

    # Format records
    for r in records:
        if 'SESSION' in r and 'SESSIONLINK' in r:
            _sess = r['SESSION']
            _link = r['SESSIONLINK']
            r['SESSION'] = f'[{_sess}]({_link})'

    # Format columns
    for i, c in enumerate(columns):
        if c['name'] == 'SESSION':
            columns[i]['type'] = 'text'
            columns[i]['presentation'] = 'markdown'
        elif _plottable(df[c['name']]):
            columns[i]['type'] = 'numeric'

    # Count how many rows are in the table
    rowcount = '{} rows'.format(len(records))

    # Return table, figure, dropdown options
    logger.debug('update_all:returning data')
    return [proc, proj, sess, records, columns, tabs, rowcount, rowcount]



import pandas as pd

from garjus import Garjus


aparc_columns = [
    'StructName',
    'NumVert',
    'SurfArea',
    'GrayVol',
    'ThickAvg',
    'ThickStd',
    'MeanCurv',
    'GausCurv',
    'FoldInd',
    'CurvInd'
]

aseg_columns = [
    'Index',
    'SegId',
    'NVoxels',
    'Volume_mm3',
    'StructName',
    'normMean',
    'normStdDev',
    'normMin',
    'normMax',
    'normRange'
]


def get_stats(xnat, fullpath):
    stats = {}

    print('downloading lh.aparc')
    xnat_file = xnat.select(fullpath).resource('SUBJ').file('stats/lh.aparc.stats').get()
    df = pd.read_table(
        xnat_file,
        comment='#',
        header=None,
        names=aparc_columns,
        sep='\s+',
        index_col='StructName'
    )

    # Left Hemisphere Thickness
    stats['caudalmiddlefrontal_lh_thickavg'] = str(df.loc['caudalmiddlefrontal', 'ThickAvg'])
    stats['entorhinal_lh_thickavg'] = str(df.loc['entorhinal', 'ThickAvg'])
    stats['lateralorbitofrontal_lh_thickavg'] = str(df.loc['lateralorbitofrontal', 'ThickAvg'])
    stats['medialorbitofrontal_lh_thickavg'] = str(df.loc['medialorbitofrontal', 'ThickAvg'])
    stats['rostralmiddlefrontal_lh_thickavg'] = str(df.loc['rostralmiddlefrontal', 'ThickAvg'])
    stats['superiorfrontal_lh_thickavg'] = str(df.loc['superiorfrontal', 'ThickAvg'])
    stats['bankssts_lh_thickavg'] = str(df.loc['bankssts', 'ThickAvg'])
    stats['caudalanteriorcingulate_lh_thickavg'] = str(df.loc['caudalanteriorcingulate', 'ThickAvg'])
    stats['cuneus_lh_thickavg'] = str(df.loc['cuneus', 'ThickAvg'])
    stats['frontalpole_lh_thickavg'] = str(df.loc['frontalpole', 'ThickAvg'])
    stats['fusiform_lh_thickavg'] = str(df.loc['fusiform', 'ThickAvg'])
    stats['inferiorparietal_lh_thickavg'] = str(df.loc['inferiorparietal', 'ThickAvg'])
    stats['inferiortemporal_lh_thickavg'] = str(df.loc['inferiortemporal', 'ThickAvg'])
    stats['insula_lh_thickavg'] = str(df.loc['insula', 'ThickAvg'])
    stats['isthmuscingulate_lh_thickavg'] = str(df.loc['isthmuscingulate', 'ThickAvg'])
    stats['lateraloccipital_lh_thickavg'] = str(df.loc['lateraloccipital', 'ThickAvg'])
    stats['lingual_lh_thickavg'] = str(df.loc['lingual', 'ThickAvg'])
    stats['middletemporal_lh_thickavg'] = str(df.loc['middletemporal', 'ThickAvg'])
    stats['parahippocampal_lh_thickavg'] = str(df.loc['parahippocampal', 'ThickAvg'])
    stats['paracentral_lh_thickavg'] = str(df.loc['paracentral', 'ThickAvg'])
    stats['parsopercularis_lh_thickavg'] = str(df.loc['parsopercularis', 'ThickAvg'])
    stats['parsorbitalis_lh_thickavg'] = str(df.loc['parsorbitalis', 'ThickAvg'])
    stats['parstriangularis_lh_thickavg'] = str(df.loc['parstriangularis', 'ThickAvg'])
    stats['pericalcarine_lh_thickavg'] = str(df.loc['pericalcarine', 'ThickAvg'])   
    stats['postcentral_lh_thickavg'] = str(df.loc['postcentral', 'ThickAvg'])
    stats['posteriorcingulate_lh_thickavg'] = str(df.loc['posteriorcingulate', 'ThickAvg'])    
    stats['precentral_lh_thickavg'] = str(df.loc['precentral', 'ThickAvg'])
    stats['precuneus_lh_thickavg'] = str(df.loc['precuneus', 'ThickAvg'])
    stats['rostralanteriorcingulate_lh_thickavg'] = str(df.loc['rostralanteriorcingulate', 'ThickAvg'])    
    stats['superiorparietal_lh_thickavg'] = str(df.loc['superiorparietal', 'ThickAvg'])
    stats['superiortemporal_lh_thickavg'] = str(df.loc['superiortemporal', 'ThickAvg'])
    stats['supramarginal_lh_thickavg'] = str(df.loc['supramarginal', 'ThickAvg'])
    stats['temporalpole_lh_thickavg'] = str(df.loc['temporalpole', 'ThickAvg'])
    stats['transversetemporal_lh_thickavg'] = str(df.loc['transversetemporal', 'ThickAvg'])

    # Left Hemisphere volumes
    stats['caudalmiddlefrontal_lh_volume'] = str(df.loc['caudalmiddlefrontal', 'GrayVol'])
    stats['entorhinal_lh_volume'] = str(df.loc['entorhinal', 'GrayVol'])
    stats['lateralorbitofrontal_lh_volume'] = str(df.loc['lateralorbitofrontal', 'GrayVol'])
    stats['medialorbitofrontal_lh_volume'] = str(df.loc['medialorbitofrontal', 'GrayVol'])
    stats['rostralmiddlefrontal_lh_volume'] = str(df.loc['rostralmiddlefrontal', 'GrayVol'])
    stats['superiorfrontal_lh_volume'] = str(df.loc['superiorfrontal', 'GrayVol'])
    stats['bankssts_lh_volume'] = str(df.loc['bankssts', 'GrayVol'])
    stats['caudalanteriorcingulate_lh_volume'] = str(df.loc['caudalanteriorcingulate', 'GrayVol'])
    stats['cuneus_lh_volume'] = str(df.loc['cuneus', 'GrayVol'])
    stats['frontalpole_lh_volume'] = str(df.loc['frontalpole', 'GrayVol'])
    stats['fusiform_lh_volume'] = str(df.loc['fusiform', 'GrayVol'])
    stats['inferiorparietal_lh_volume'] = str(df.loc['inferiorparietal', 'GrayVol'])
    stats['inferiortemporal_lh_volume'] = str(df.loc['inferiortemporal', 'GrayVol'])
    stats['insula_lh_volume'] = str(df.loc['insula', 'GrayVol'])
    stats['isthmuscingulate_lh_volume'] = str(df.loc['isthmuscingulate', 'GrayVol'])
    stats['lateraloccipital_lh_volume'] = str(df.loc['lateraloccipital', 'GrayVol'])
    stats['lingual_lh_volume'] = str(df.loc['lingual', 'GrayVol'])
    stats['middletemporal_lh_volume'] = str(df.loc['middletemporal', 'GrayVol'])
    stats['parahippocampal_lh_volume'] = str(df.loc['parahippocampal', 'GrayVol'])
    stats['paracentral_lh_volume'] = str(df.loc['paracentral', 'GrayVol'])
    stats['parsopercularis_lh_volume'] = str(df.loc['parsopercularis', 'GrayVol'])
    stats['parsorbitalis_lh_volume'] = str(df.loc['parsorbitalis', 'GrayVol'])
    stats['parstriangularis_lh_volume'] = str(df.loc['parstriangularis', 'GrayVol'])
    stats['pericalcarine_lh_volume'] = str(df.loc['pericalcarine', 'GrayVol'])   
    stats['postcentral_lh_volume'] = str(df.loc['postcentral', 'GrayVol'])
    stats['posteriorcingulate_lh_volume'] = str(df.loc['posteriorcingulate', 'GrayVol'])    
    stats['precentral_lh_volume'] = str(df.loc['precentral', 'GrayVol'])
    stats['precuneus_lh_volume'] = str(df.loc['precuneus', 'GrayVol'])
    stats['rostralanteriorcingulate_lh_volume'] = str(df.loc['rostralanteriorcingulate', 'GrayVol'])    
    stats['superiorparietal_lh_volume'] = str(df.loc['superiorparietal', 'GrayVol'])
    stats['superiortemporal_lh_volume'] = str(df.loc['superiortemporal', 'GrayVol'])
    stats['supramarginal_lh_volume'] = str(df.loc['supramarginal', 'GrayVol'])
    stats['temporalpole_lh_volume'] = str(df.loc['temporalpole', 'GrayVol'])
    stats['transversetemporal_lh_volume'] = str(df.loc['transversetemporal', 'GrayVol'])

    # Left Hemisphere Surface Area
    stats['caudalmiddlefrontal_lh_surfarea'] = str(df.loc['caudalmiddlefrontal', 'SurfArea'])
    stats['entorhinal_lh_surfarea'] = str(df.loc['entorhinal', 'SurfArea'])
    stats['lateralorbitofrontal_lh_surfarea'] = str(df.loc['lateralorbitofrontal', 'SurfArea'])
    stats['medialorbitofrontal_lh_surfarea'] = str(df.loc['medialorbitofrontal', 'SurfArea'])
    stats['rostralmiddlefront_lh_surfarea'] = str(df.loc['rostralmiddlefrontal', 'SurfArea'])
    stats['superiorfrontal_lh_surfarea'] = str(df.loc['superiorfrontal', 'SurfArea'])
    stats['bankssts_lh_surfarea'] = str(df.loc['bankssts', 'SurfArea'])
    stats['caudalanteriorcingulate_lh_surfarea'] = str(df.loc['caudalanteriorcingulate', 'SurfArea'])
    stats['cuneus_lh_surfarea'] = str(df.loc['cuneus', 'SurfArea'])
    stats['frontalpole_lh_surfarea'] = str(df.loc['frontalpole', 'SurfArea'])
    stats['fusiform_lh_surfarea'] = str(df.loc['fusiform', 'SurfArea'])
    stats['inferiorparietal_lh_surfarea'] = str(df.loc['inferiorparietal', 'SurfArea'])
    stats['inferiortemporal_lh_surfarea'] = str(df.loc['inferiortemporal', 'SurfArea'])
    stats['insula_lh_surfarea'] = str(df.loc['insula', 'SurfArea'])
    stats['isthmuscingulate_lh_surfarea'] = str(df.loc['isthmuscingulate', 'SurfArea'])
    stats['lateraloccipital_lh_surfarea'] = str(df.loc['lateraloccipital', 'SurfArea'])
    stats['lingual_lh_surfarea'] = str(df.loc['lingual', 'SurfArea'])
    stats['middletemporal_lh_surfarea'] = str(df.loc['middletemporal', 'SurfArea'])
    stats['parahippocampal_lh_surfarea'] = str(df.loc['parahippocampal', 'SurfArea'])
    stats['paracentral_lh_surfarea'] = str(df.loc['paracentral', 'SurfArea'])
    stats['parsopercularis_lh_surfarea'] = str(df.loc['parsopercularis', 'SurfArea'])
    stats['parsorbitalis_lh_surfarea'] = str(df.loc['parsorbitalis', 'SurfArea'])
    stats['parstriangularis_lh_surfarea'] = str(df.loc['parstriangularis', 'SurfArea'])
    stats['pericalcarine_lh_surfarea'] = str(df.loc['pericalcarine', 'SurfArea'])   
    stats['postcentral_lh_surfarea'] = str(df.loc['postcentral', 'SurfArea'])
    stats['posteriorcingulate_lh_surfarea'] = str(df.loc['posteriorcingulate', 'SurfArea'])    
    stats['precentral_lh_surfarea'] = str(df.loc['precentral', 'SurfArea'])
    stats['precuneus_lh_surfarea'] = str(df.loc['precuneus', 'SurfArea'])
    stats['rostralanteriorcingulate_lh_surfarea'] = str(df.loc['rostralanteriorcingulate', 'SurfArea'])    
    stats['superiorparietal_lh_surfarea'] = str(df.loc['superiorparietal', 'SurfArea'])
    stats['superiortemporal_lh_surfarea'] = str(df.loc['superiortemporal', 'SurfArea'])
    stats['supramarginal_lh_surfarea'] = str(df.loc['supramarginal', 'SurfArea'])
    stats['temporalpole_lh_surfarea'] = str(df.loc['temporalpole', 'SurfArea'])
    stats['transversetemporal_lh_surfarea'] = str(df.loc['transversetemporal', 'SurfArea'])

    print('downloading rh.aparc')
    xnat_file = xnat.select(fullpath).resource('SUBJ').file('stats/rh.aparc.stats').get()
    df = pd.read_table(
        xnat_file,
        comment='#',
        header=None,
        names=aparc_columns,
        sep='\s+',
        index_col='StructName'
    )

    # Right Hemisphere Thickness Average
    stats['caudalmiddlefrontal_rh_thickavg'] = str(df.loc['caudalmiddlefrontal', 'ThickAvg'])
    stats['entorhinal_rh_thickavg'] = str(df.loc['entorhinal', 'ThickAvg'])
    stats['lateralorbitofrontal_rh_thickavg'] = str(df.loc['lateralorbitofrontal', 'ThickAvg'])
    stats['medialorbitofrontal_rh_thickavg'] = str(df.loc['medialorbitofrontal', 'ThickAvg'])
    stats['rostralmiddlefrontal_rh_thickavg'] = str(df.loc['rostralmiddlefrontal', 'ThickAvg'])
    stats['superiorfrontal_rh_thickavg'] = str(df.loc['superiorfrontal', 'ThickAvg'])
    stats['bankssts_rh_thickavg'] = str(df.loc['bankssts', 'ThickAvg'])
    stats['caudalanteriorcingulate_rh_thickavg'] = str(df.loc['caudalanteriorcingulate', 'ThickAvg'])
    stats['cuneus_rh_thickavg'] = str(df.loc['cuneus', 'ThickAvg'])
    stats['frontalpole_rh_thickavg'] = str(df.loc['frontalpole', 'ThickAvg'])
    stats['fusiform_rh_thickavg'] = str(df.loc['fusiform', 'ThickAvg'])
    stats['inferiorparietal_rh_thickavg'] = str(df.loc['inferiorparietal', 'ThickAvg'])
    stats['inferiortemporal_rh_thickavg'] = str(df.loc['inferiortemporal', 'ThickAvg'])
    stats['insula_rh_thickavg'] = str(df.loc['insula', 'ThickAvg'])
    stats['isthmuscingulate_rh_thickavg'] = str(df.loc['isthmuscingulate', 'ThickAvg'])
    stats['lateraloccipital_rh_thickavg'] = str(df.loc['lateraloccipital', 'ThickAvg'])
    stats['lingual_rh_thickavg'] = str(df.loc['lingual', 'ThickAvg'])
    stats['middletemporal_rh_thickavg'] = str(df.loc['middletemporal', 'ThickAvg'])
    stats['parahippocampal_rh_thickavg'] = str(df.loc['parahippocampal', 'ThickAvg'])
    stats['paracentral_rh_thickavg'] = str(df.loc['paracentral', 'ThickAvg'])
    stats['parsopercularis_rh_thickavg'] = str(df.loc['parsopercularis', 'ThickAvg'])
    stats['parsorbitalis_rh_thickavg'] = str(df.loc['parsorbitalis', 'ThickAvg'])
    stats['parstriangularis_rh_thickavg'] = str(df.loc['parstriangularis', 'ThickAvg'])
    stats['pericalcarine_rh_thickavg'] = str(df.loc['pericalcarine', 'ThickAvg'])   
    stats['postcentral_rh_thickavg'] = str(df.loc['postcentral', 'ThickAvg'])
    stats['posteriorcingulate_rh_thickavg'] = str(df.loc['posteriorcingulate', 'ThickAvg'])    
    stats['precentral_rh_thickavg'] = str(df.loc['precentral', 'ThickAvg'])
    stats['precuneus_rh_thickavg'] = str(df.loc['precuneus', 'ThickAvg'])
    stats['rostralanteriorcingulate_rh_thickavg'] = str(df.loc['rostralanteriorcingulate', 'ThickAvg'])    
    stats['superiorparietal_rh_thickavg'] = str(df.loc['superiorparietal', 'ThickAvg'])
    stats['superiortemporal_rh_thickavg'] = str(df.loc['superiortemporal', 'ThickAvg'])
    stats['supramarginal_rh_thickavg'] = str(df.loc['supramarginal', 'ThickAvg'])
    stats['temporalpole_rh_thickavg'] = str(df.loc['temporalpole', 'ThickAvg'])
    stats['transversetemporal_rh_thickavg'] = str(df.loc['transversetemporal', 'ThickAvg'])

    # Right Hemisphere Volumes
    stats['caudalmiddlefrontal_rh_volume'] = str(df.loc['caudalmiddlefrontal', 'GrayVol'])
    stats['entorhinal_rh_volume'] = str(df.loc['entorhinal', 'GrayVol'])
    stats['lateralorbitofrontal_rh_volume'] = str(df.loc['lateralorbitofrontal', 'GrayVol'])
    stats['medialorbitofrontal_rh_volume'] = str(df.loc['medialorbitofrontal', 'GrayVol'])
    stats['rostralmiddlefrontal_rh_volume'] = str(df.loc['rostralmiddlefrontal', 'GrayVol'])
    stats['superiorfrontal_rh_volume'] = str(df.loc['superiorfrontal', 'GrayVol'])
    stats['bankssts_rh_volume'] = str(df.loc['bankssts', 'GrayVol'])
    stats['caudalanteriorcingulate_rh_volume'] = str(df.loc['caudalanteriorcingulate', 'GrayVol'])
    stats['cuneus_rh_volume'] = str(df.loc['cuneus', 'GrayVol'])
    stats['frontalpole_rh_volume'] = str(df.loc['frontalpole', 'GrayVol'])
    stats['fusiform_rh_volume'] = str(df.loc['fusiform', 'GrayVol'])
    stats['inferiorparietal_rh_volume'] = str(df.loc['inferiorparietal', 'GrayVol'])
    stats['inferiortemporal_rh_volume'] = str(df.loc['inferiortemporal', 'GrayVol'])
    stats['insula_rh_volume'] = str(df.loc['insula', 'GrayVol'])
    stats['isthmuscingulate_rh_volume'] = str(df.loc['isthmuscingulate', 'GrayVol'])
    stats['lateraloccipital_rh_volume'] = str(df.loc['lateraloccipital', 'GrayVol'])
    stats['lingual_rh_volume'] = str(df.loc['lingual', 'GrayVol'])
    stats['middletemporal_rh_volume'] = str(df.loc['middletemporal', 'GrayVol'])
    stats['parahippocampal_rh_volume'] = str(df.loc['parahippocampal', 'GrayVol'])
    stats['paracentral_rh_volume'] = str(df.loc['paracentral', 'GrayVol'])
    stats['parsopercularis_rh_volume'] = str(df.loc['parsopercularis', 'GrayVol'])
    stats['parsorbitalis_rh_volume'] = str(df.loc['parsorbitalis', 'GrayVol'])
    stats['parstriangularis_rh_volume'] = str(df.loc['parstriangularis', 'GrayVol'])
    stats['pericalcarine_rh_volume'] = str(df.loc['pericalcarine', 'GrayVol'])   
    stats['postcentral_rh_volume'] = str(df.loc['postcentral', 'GrayVol'])
    stats['posteriorcingulate_rh_volume'] = str(df.loc['posteriorcingulate', 'GrayVol'])    
    stats['precentral_rh_volume'] = str(df.loc['precentral', 'GrayVol'])
    stats['precuneus_rh_volume'] = str(df.loc['precuneus', 'GrayVol'])
    stats['rostralanteriorcingulate_rh_volume'] = str(df.loc['rostralanteriorcingulate', 'GrayVol'])    
    stats['superiorparietal_rh_volume'] = str(df.loc['superiorparietal', 'GrayVol'])
    stats['superiortemporal_rh_volume'] = str(df.loc['superiortemporal', 'GrayVol'])
    stats['supramarginal_rh_volume'] = str(df.loc['supramarginal', 'GrayVol'])
    stats['temporalpole_rh_volume'] = str(df.loc['temporalpole', 'GrayVol'])
    stats['transversetemporal_rh_volume'] = str(df.loc['transversetemporal', 'GrayVol'])

    # Right Hemisphere Surface Area
    stats['caudalmiddlefrontal_rh_surfarea'] = str(df.loc['caudalmiddlefrontal', 'SurfArea'])
    stats['entorhinal_rh_surfarea'] = str(df.loc['entorhinal', 'SurfArea'])
    stats['lateralorbitofrontal_rh_surfarea'] = str(df.loc['lateralorbitofrontal', 'SurfArea'])
    stats['medialorbitofrontal_rh_surfarea'] = str(df.loc['medialorbitofrontal', 'SurfArea'])
    stats['rostralmiddlefront_rh_surfarea'] = str(df.loc['rostralmiddlefrontal', 'SurfArea'])
    stats['superiorfrontal_rh_surfarea'] = str(df.loc['superiorfrontal', 'SurfArea'])
    stats['bankssts_rh_surfarea'] = str(df.loc['bankssts', 'SurfArea'])
    stats['caudalanteriorcingulate_rh_surfarea'] = str(df.loc['caudalanteriorcingulate', 'SurfArea'])
    stats['cuneus_rh_surfarea'] = str(df.loc['cuneus', 'SurfArea'])
    stats['frontalpole_rh_surfarea'] = str(df.loc['frontalpole', 'SurfArea'])
    stats['fusiform_rh_surfarea'] = str(df.loc['fusiform', 'SurfArea'])
    stats['inferiorparietal_rh_surfarea'] = str(df.loc['inferiorparietal', 'SurfArea'])
    stats['inferiortemporal_rh_surfarea'] = str(df.loc['inferiortemporal', 'SurfArea'])
    stats['insula_rh_surfarea'] = str(df.loc['insula', 'SurfArea'])
    stats['isthmuscingulate_rh_surfarea'] = str(df.loc['isthmuscingulate', 'SurfArea'])
    stats['lateraloccipital_rh_surfarea'] = str(df.loc['lateraloccipital', 'SurfArea'])
    stats['lingual_rh_surfarea'] = str(df.loc['lingual', 'SurfArea'])
    stats['middletemporal_rh_surfarea'] = str(df.loc['middletemporal', 'SurfArea'])
    stats['parahippocampal_rh_surfarea'] = str(df.loc['parahippocampal', 'SurfArea'])
    stats['paracentral_rh_surfarea'] = str(df.loc['paracentral', 'SurfArea'])
    stats['parsopercularis_rh_surfarea'] = str(df.loc['parsopercularis', 'SurfArea'])
    stats['parsorbitalis_rh_surfarea'] = str(df.loc['parsorbitalis', 'SurfArea'])
    stats['parstriangularis_rh_surfarea'] = str(df.loc['parstriangularis', 'SurfArea'])
    stats['pericalcarine_rh_surfarea'] = str(df.loc['pericalcarine', 'SurfArea'])   
    stats['postcentral_rh_surfarea'] = str(df.loc['postcentral', 'SurfArea'])
    stats['posteriorcingulate_rh_surfarea'] = str(df.loc['posteriorcingulate', 'SurfArea'])    
    stats['precentral_rh_surfarea'] = str(df.loc['precentral', 'SurfArea'])
    stats['precuneus_rh_surfarea'] = str(df.loc['precuneus', 'SurfArea'])
    stats['rostralanteriorcingulate_rh_surfarea'] = str(df.loc['rostralanteriorcingulate', 'SurfArea'])    
    stats['superiorparietal_rh_surfarea'] = str(df.loc['superiorparietal', 'SurfArea'])
    stats['superiortemporal_rh_surfarea'] = str(df.loc['superiortemporal', 'SurfArea'])
    stats['supramarginal_rh_surfarea'] = str(df.loc['supramarginal', 'SurfArea'])
    stats['temporalpole_rh_surfarea'] = str(df.loc['temporalpole', 'SurfArea'])
    stats['transversetemporal_rh_surfarea'] = str(df.loc['transversetemporal', 'SurfArea'])

    print('downloading aseg')
    xnat_file = xnat.select(fullpath).resource('SUBJ').file('stats/aseg.stats').get()
    df = pd.read_table(
        xnat_file,
        comment='#',
        header=None,
        names=aseg_columns,
        sep='\s+',
        index_col='StructName'
    )

    # Subcortical Volumes Left
    stats['accumbensarea_lh_volume'] = str(df.loc['Left-Accumbens-area', 'Volume_mm3'])
    stats['amygdala_lh_volume'] = str(df.loc['Left-Amygdala', 'Volume_mm3'])
    stats['caudate_lh_volume'] = str(df.loc['Left-Caudate', 'Volume_mm3'])
    stats['hippocampus_lh_volume'] = str(df.loc['Left-Hippocampus', 'Volume_mm3'])
    stats['pallidum_lh_volume'] = str(df.loc['Left-Pallidum', 'Volume_mm3'])
    stats['putamen_lh_volume'] = str(df.loc['Left-Putamen', 'Volume_mm3'])
    stats['thalamus_lh_volume'] = str(df.loc['Left-Thalamus', 'Volume_mm3'])
    stats['ventraldc_lh_volume'] = str(df.loc['Left-VentralDC', 'Volume_mm3'])

    # Subcortical Volumes Right
    stats['accumbensarea_rh_volume'] = str(df.loc['Right-Accumbens-area', 'Volume_mm3'])
    stats['amygdala_rh_volume'] = str(df.loc['Right-Amygdala', 'Volume_mm3'])
    stats['caudate_rh_volume'] = str(df.loc['Right-Caudate', 'Volume_mm3'])
    stats['hippocampus_rh_volume'] = str(df.loc['Right-Hippocampus', 'Volume_mm3'])
    stats['pallidum_rh_volume'] = str(df.loc['Right-Pallidum', 'Volume_mm3'])
    stats['putamen_rh_volume'] = str(df.loc['Right-Putamen', 'Volume_mm3'])
    stats['thalamus_rh_volume'] = str(df.loc['Right-Thalamus', 'Volume_mm3'])
    stats['ventraldc_rh_volume'] = str(df.loc['Right-VentralDC', 'Volume_mm3'])

    # Hypointensities
    stats['wmh_volume'] = str(df.loc['WM-hypointensities', 'Volume_mm3'])

    return stats



g = Garjus()

df = g.assessors(projects=['REMBRANDT'])

df = df[df.PROCTYPE == 'FS7_v1']

df = df.sort_values('ASSR')

for i, row in df.iterrows():
    print(i, row['full_path'])
    stats = get_stats(g.xnat(), row['full_path'])
    print('set stats')
    g.set_stats(row['PROJECT'], row['SUBJECT'], row['SESSION'], row['ASSR'], stats)

print('DONE!')



import os
import pandas as pd


data_files = [
    '2023-06-19 #1.csv',
    '2023-10-04 11.34.26 Assessment Data.xls',
    '2024-02-13 12.30.50 Assessment Data - Emily 2.csv',
    '2023-10-04 11.48.57 Assessment Data.xls',
    '2024-02-13 12.46.04 Assessment Data - Devika 3.csv',
    '2023-06-19 10.06.16 Assessment Data.csv',
    '2024-02-13 12.30.13 Assessment Data.csv'
]

score_files = [
    '2023-06-19 #2.csv',
    '2023-06-19 #4.xlsm',
    '2023-06-19 #5.xlsm',
    '2023-10-04 11.34.26 Assessment Scores.xls',
    '2024-02-13 12.30.13 Assessment Scores.csv',
    '2024-02-13 12.46.04 Assessment Scores - Devika2.csv',
    '2023-06-19 10.06.16 Assessment Scores.csv',
    '2022-12-20 chris Assessment Scores.csv',
    '2023-10-04 11.48.57 Assessment Scores.xls',
    '2024-02-13 12.30.50 Assessment Scores - Emily 1.csv',
    '2022-12-20 devika Assessment Scores.csv', 
]

# Load Reg Data
reg1 = pd.read_csv('/home/boydb1/Neurocog Data/2024-02-13 12.30.13 Registration Data.csv', dtype={'PIN': str})
reg1 = reg1.set_index(['PIN', 'Assessment Name'])

reg2 = pd.read_csv('/home/boydb1/Neurocog Data/2024-02-13 12.30.50 Registration Data - Emily3.csv', dtype={'PIN': str})
reg2 = reg2.set_index(['PIN', 'Assessment Name'])

reg3 = pd.read_csv('/home/boydb1/Neurocog Data/2024-02-13 12.46.04 Registration Data - Devika1.csv',  dtype={'PIN': str})
reg3 = reg3.set_index(['PIN', 'Assessment Name'])

reg4 = pd.read_excel('/home/boydb1/Neurocog Data/2023-10-04 11.34.26 Registration Data.xls', dtype={'PIN': str})
reg4 = reg4.set_index(['PIN', 'Assessment Name'])

reg4 = pd.read_excel('/home/boydb1/Neurocog Data/2023-10-04 11.34.26 Registration Data.xls', dtype={'PIN': str})
reg4 = reg4.set_index(['PIN', 'Assessment Name'])

reg5 = pd.read_excel('/home/boydb1/Neurocog Data/2023-10-04 11.48.57 Registration Data.xls', dtype={'PIN': str})
reg5 = reg5.set_index(['PIN', 'Assessment Name'])

reg8 = pd.read_excel('/home/boydb1/Neurocog Data/2023-06-19 #6.xlsm', dtype={'PIN': str})
reg8 = reg8.set_index(['PIN', 'Assessment Name'])

reg9 = pd.read_csv('/home/boydb1/Neurocog Data/2023-06-19 10.06.16 Registration Data.csv', dtype={'PIN': str})
reg9 = reg9.set_index(['PIN', 'Assessment Name'])

# Merge Reg Data
dfr = pd.concat([reg1, reg2, reg3, reg4, reg5, reg8]) #, verify_integrity=True)
dfr = dfr.drop_duplicates()

# Save Reg Data
for i, row in dfr.reset_index().iterrows():
    if not row['PIN'].isnumeric():
        continue

    filename = f'/home/boydb1/Nair_Neurocog/{row["PIN"]}_{row["Assessment Name"]}_Reg.csv'.replace(' ', '')
    print(filename)
    pd.DataFrame([row]).to_csv(filename, index=False)

# Load data files to dfd DataFrame
dfd = pd.DataFrame()
for f in data_files:
    print(f)
    if f.endswith('.csv'):
        reg = pd.read_csv(f'/home/boydb1/Neurocog Data/{f}', dtype={'PIN': str})
    elif f.endswith('.xls'):
        reg = pd.read_excel(f'/home/boydb1/Neurocog Data/{f}', dtype={'PIN': str})

    dfd = pd.concat([dfd, reg])

# Save Data for each PIN/Assessment
visits = dfd[['PIN', 'Assessment Name']].drop_duplicates().sort_values(['PIN', 'Assessment Name'])
for i, v in visits.iterrows():
    if not v['PIN'].isnumeric():
        print(f'not numeric:{i}:{v["PIN"]}')
        continue

    filename = f'/home/boydb1/Nair_Neurocog/{v["PIN"]}_{v["Assessment Name"]}_Data.csv'.replace(' ', '')
    vdfd = dfd[(dfd['PIN'] == v['PIN']) & (dfd['Assessment Name'] == v['Assessment Name'])]
    print(v['PIN'], v['Assessment Name'], len(vdfd), filename)
    pd.DataFrame(vdfd).to_csv(filename, index=False)

# Load score data
dfs = pd.DataFrame()
for f in score_files:
    print(f)
    if f.endswith('.csv'):
        df = pd.read_csv(f'/home/boydb1/Neurocog Data/{f}', dtype={'PIN': str})
    elif f.endswith('.xls'):
        df = pd.read_excel(f'/home/boydb1/Neurocog Data/{f}', dtype={'PIN': str})

    dfs = pd.concat([dfs, df])

# Save score data
score_visits = dfs[['PIN', 'Assessment Name']].drop_duplicates().sort_values(['PIN', 'Assessment Name'])
for i, v in score_visits.iterrows():
    if not v['PIN'].isnumeric():
        continue

    filename = f'/home/boydb1/Nair_Neurocog/{v["PIN"]}_{v["Assessment Name"]}_Score.csv'.replace(' ', '')
    vdfs = dfs[(dfs['PIN'] == v['PIN']) & (dfs['Assessment Name'] == v['Assessment Name'])]
    print(v['PIN'], v['Assessment Name'], len(vdfs), filename)
    pd.DataFrame(vdfs).to_csv(filename, index=False)

print('DONE!')
